# -*- coding: utf-8 -*-
r"""
This module contains the QCD beta function coefficients.

See :doc:`pQCD ingredients </theory/pQCD>`.
"""

import numba as nb

from . import constants
from .harmonics.constants import zeta3


@nb.njit(cache=True)
def beta_qcd_as2(nf):
    """Computes the first coefficient of the QCD beta function.

    Implements Eq. (3.1) of :cite:`Herzog:2017ohr`.

    Parameters
    ----------
        nf : int
            number of active flavors

    Returns
    -------
        beta_qcd_as2 : float
            first coefficient of the QCD beta function :math:`\\beta_qcd_as2^{n_f}`

    """
    beta_qcd_as2 = 11.0 / 3.0 * constants.CA - 4.0 / 3.0 * constants.TR * nf
    return beta_qcd_as2


@nb.njit(cache=True)
def beta_qed_aem2(nf):
    """Computes the first coefficient of the QED beta function.

    Implements Eq. (7) of :cite:`Surguladze:1996hx`.

    Parameters
    ----------
        nf : int
            number of active flavors

    Returns
    -------
        beta_qed_aem2 : float
            first coefficient of the QED beta function :math:`\\beta_qed_aem2^{n_f}`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    beta_qed_aem2 = -4.0 / 3 * constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    return beta_qed_aem2


@nb.njit(cache=True)
def beta_qcd_as3(nf):
    """Computes the second coefficient of the QCD beta function.

    Implements Eq. (3.2) of :cite:`Herzog:2017ohr`.

    Parameters
    ----------
        nf : int
            number of active flavors

    Returns
    -------
        beta_qcd_as3 : float
            second coefficient of the QCD beta function :math:`\\beta_qcd_as3^{n_f}`

    """
    TF = constants.TR * nf
    b_ca2 = 34.0 / 3.0 * constants.CA * constants.CA
    b_ca = -20.0 / 3.0 * constants.CA * TF
    b_cf = -4.0 * constants.CF * TF
    beta_qcd_as3 = b_ca2 + b_ca + b_cf
    return beta_qcd_as3


@nb.njit(cache=True)
def beta_qed_aem3(nf):
    """Computes the second coefficient of the QED beta function.

    Implements Eq. (7) of :cite:`Surguladze:1996hx`.

    Parameters
    ----------
        nf : int
            number of active flavors

    Returns
    -------
        beta_qed_aem3 : float
            second coefficient of the QED beta function :math:`\\beta_qed_aem3^{n_f}`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    beta_qed_aem3 = (
        -4.0 * constants.NC * (nu * constants.eu2**2 + nd * constants.ed2**2)
    )
    return beta_qed_aem3


@nb.njit(cache=True)
def beta_qcd_as2aem1(nf):
    """Computes the first QED correction of the QCD beta function.

    Implements Eq. (7) of :cite:`Surguladze:1996hx`.

    Parameters
    ----------
        nf : int
            number of active flavors

    Returns
    -------
        beta_as2aem1 : float
            first QED correction of the QCD beta function :math:`\\beta_as2aem1^{n_f}`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    beta_qcd_as2aem1 = -4.0 * constants.TR * (nu * constants.eu2 + nd * constants.ed2)
    return beta_qcd_as2aem1


@nb.njit(cache=True)
def beta_qed_aem2as1(nf):
    """Computes the first QCD correction of the QED beta function.

    Implements Eq. (7) of :cite:`Surguladze:1996hx`.

    Parameters
    ----------
        nf : int
            number of active flavors

    Returns
    -------
        beta_aem2as1 : float
            first QCD correction of the QED beta function :math:`\\beta_aem2as1^{n_f}`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    beta_qed_aem2as1 = (
        -4.0 * constants.CF * constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    )
    return beta_qed_aem2as1


@nb.njit(cache=True)
def beta_qcd_as4(nf):
    """Computes the third coefficient of the QCD beta function

    Implements Eq. (3.3) of :cite:`Herzog:2017ohr`.

    Parameters
    ----------
        nf : int
            number of active flavors

    Returns
    -------
        beta_qcd_as4 : float
            third coefficient of the QCD beta function :math:`\\beta_qcd_as4^{n_f}`

    """
    TF = constants.TR * nf
    beta_qcd_as4 = (
        2857.0 / 54.0 * constants.CA * constants.CA * constants.CA
        - 1415.0 / 27.0 * constants.CA * constants.CA * TF
        - 205.0 / 9.0 * constants.CF * constants.CA * TF
        + 2.0 * constants.CF * constants.CF * TF
        + 44.0 / 9.0 * constants.CF * TF * TF
        + 158.0 / 27.0 * constants.CA * TF * TF
    )
    return beta_qcd_as4


@nb.njit(cache=True)
def beta_qcd_as5(nf):
    """Computes the fourth coefficient of the QCD beta function

    Implements Eq. (3.6) of :cite:`Herzog:2017ohr`.

    Parameters
    ----------
        nf : int
            number of active flavors

    Returns
    -------
        beta_qcd_as5 : float
            fourth coefficient of the QCD beta function :math:`\\beta_qcd_as5^{n_f}`

    """
    beta_qcd_as5 = (
        149753.0 / 6.0
        + 3564.0 * zeta3
        + nf * (-1078361.0 / 162.0 - 6508.0 / 27.0 * zeta3)
        + nf**2 * (50065.0 / 162.0 + 6472.0 / 81.0 * zeta3)
        + 1093.0 / 729.0 * nf**3
    )
    return beta_qcd_as5


@nb.njit(cache=True)
def beta_qcd(k, nf):
    """Compute value of a beta_qcd coefficients

    Parameters
    ----------
        k : tuple(int, int)
            perturbative order
        nf : int
            number of active flavors

    Returns
    -------
        beta_qcd : float
            beta_qcd_k(nf)

    """
    beta_ = 0
    if k == (2, 0):
        beta_ = beta_qcd_as2(nf)
    elif k == (3, 0):
        beta_ = beta_qcd_as3(nf)
    elif k == (4, 0):
        beta_ = beta_qcd_as4(nf)
    elif k == (5, 0):
        beta_ = beta_qcd_as5(nf)
    elif k == (2, 1):
        beta_ = beta_qcd_as2aem1(nf)
    else:
        raise ValueError("Beta_QCD coefficients beyond N3LO are not implemented!")
    return beta_


@nb.njit(cache=True)
def beta_qed(k, nf):
    """Compute value of a beta_qed coefficients

    Parameters
    ----------
        k : tuple(int, int)
            perturbative order
        nf : int
            number of active flavors

    Returns
    -------
        beta_qed : float
            beta_qed_k(nf)

    """
    beta_ = 0
    if k == (0, 2):
        beta_ = beta_qed_aem2(nf)
    elif k == (0, 3):
        beta_ = beta_qed_aem3(nf)
    elif k == (1, 2):
        beta_ = beta_qed_aem2as1(nf)
    else:
        raise ValueError("Beta_QED coefficients beyond NLO are not implemented!")
    return beta_


@nb.njit(cache=True)
def b_qcd(k, nf):
    """Compute b_qcd coefficient.

    Parameters
    ----------
        k : tuple(int, int)
            perturbative order
        nf : int
            number of active flavors

    Returns
    -------
        b_qcd : float
            b_qcd_k(nf)

    """
    return beta_qcd(k, nf) / beta_qcd((2, 0), nf)


@nb.njit(cache=True)
def b_qed(k, nf):
    """Compute b_qed coefficient.

    Parameters
    ----------
        k : tuple(int, int)
            perturbative order
        nf : int
            number of active flavors

    Returns
    -------
        b_qed : float
            b_qed_k(nf)

    """
    return beta_qed(k, nf) / beta_qed((0, 2), nf)



# -*- coding: utf-8 -*-
"""
Library providing all necessary tools for PDF interpolation.

This library also provides a class to generate the interpolator :class:`InterpolatorDispatcher`.
Upon construction the dispatcher generates a number of functions
to evaluate the interpolator.
"""
import logging
import math

import numba as nb
import numpy as np
import scipy.special as sp

logger = logging.getLogger(__name__)


class Area:
    """
    Class that define each of the area
    of each of the subgrid interpolators.

    Upon construction an array of coefficients
    is generated.

    Parameters
    ----------
        lower_index: int
            lower index of the area
        poly_number: int
            number of polynomial
        block: tuple(int, int)
            kmin and kmax
        xgrid: numpy.ndarray
            Grid in x-space from which the interpolators are constructed
    """

    def __init__(self, lower_index, poly_number, block, xgrid):
        # check range
        if poly_number < block[0] or block[1] < poly_number:
            raise ValueError(
                f"polynom #{poly_number} cannot be a part of the block which"
                f"spans from {block[0]} to {block[1]}"
            )
        self.xmin = xgrid[lower_index]
        self.xmax = xgrid[lower_index + 1]
        self.poly_number = poly_number
        self.kmin = block[0]
        self.kmax = block[1]
        self.coefs = self._compute_coefs(xgrid)

    def _reference_indices(self):
        """Iterate over all indices which are part of the block"""
        for k in range(self.kmin, self.kmax + 1):
            if k != self.poly_number:
                yield k

    def _compute_coefs(self, xgrid):
        """Compute the coefficients for this area given a grid on x"""
        denominator = 1.0
        coeffs = np.ones(1)
        xj = xgrid[self.poly_number]
        for s, k in enumerate(self._reference_indices()):
            xk = xgrid[k]
            denominator *= xj - xk
            Mxk_coeffs = -xk * coeffs
            coeffs = np.concatenate(([0.0], coeffs))
            coeffs[: s + +1] += Mxk_coeffs
        coeffs /= denominator
        return coeffs

    def __iter__(self):
        """Iterates the generated coefficients"""
        for coef in self.coefs:
            yield coef


@nb.njit(cache=True)
def log_evaluate_Nx(N, logx, area_list):
    r"""
    Evaluates a single logarithmic Lagrange interpolator in N-space multiplied
    by the Mellin-inversion factor.

    .. math::
        \tilde p(N)*\exp(- N * \ln(x))

    Parameters
    ----------
        N : complex
            Mellin variable
        logx : float
            logarithm of inversion point
        area_list : list
            area configuration of basis function

    Returns
    -------
        res : float
            kernel * inversion factor
    """
    res = 0.0
    for a in area_list:
        logxmin = a[0]
        logxmax = a[1]
        coefs = a[2:]
        # skip area completely?
        if logx >= logxmax:
            continue
        umax = N * logxmax
        umin = N * logxmin
        emax = np.exp(N * (logxmax - logx))
        emin = np.exp(N * (logxmin - logx))
        for i, coef in enumerate(coefs):
            tmp = 0.0
            facti = math.gamma(i + 1) * pow(-1, i) / pow(N, i + 1)
            for k in range(i + 1):
                factk = 1.0 / math.gamma(k + 1)
                # this condition is actually not necessary in python since
                # there pow(0,0) == 1 and apparently this is inherited in
                # Numba/C, however this is mathematically cleaner
                if umax == 0.0 and k == 0:
                    pmax = emax
                else:
                    pmax = pow(-umax, k) * emax
                # drop factor by analytics?
                if logx >= logxmin:
                    pmin = 0
                else:
                    pmin = pow(-umin, k) * emin
                tmp += factk * (pmax - pmin)
            res += coef * facti * tmp
    return res


@nb.njit(cache=True)
def evaluate_Nx(N, logx, area_list):
    r"""
    Evaluates a single linear Lagrange interpolator in N-space multiplied
    by the Mellin-inversion factor.

    .. math::
        \tilde p(N)*\exp(- N * \ln(x))

    Parameters
    ----------
        N : complex
            Mellin variable
        logx : float
            logarithm of inversion point
        area_list : list
            area configuration of basis function

    Returns
    -------
        res : float
            basis function * inversion factor
    """
    res = 0.0
    for a in area_list:
        xmin = a[0]
        xmax = a[1]
        coefs = a[2:]
        lnxmax = np.log(xmax)
        # skip area completely?
        if logx >= lnxmax:
            continue
        for i, coef in enumerate(coefs):
            if xmin == 0.0:
                low = 0.0
            else:
                lnxmin = np.log(xmin)
                low = np.exp(N * (lnxmin - logx) + i * lnxmin)
            up = np.exp(N * (lnxmax - logx) + i * lnxmax)
            res += coef * (up - low) / (N + i)
    return res


@nb.njit(cache=True)
def evaluate_grid(N, is_log, logx, area_list):
    """
    Evaluate interpolator in N-space

    Parameters
    ----------
        N : complex
            Mellin variable
        is_log : boolean
            is a logarithmic interpolation
        logx : float
            logarithm of inversion point
        area_list : list
            area configuration of basis function

    Returns
    -------
        pj : float
            basis function * inversion factor
    """ ""
    if is_log:
        pj = log_evaluate_Nx(N, logx, area_list)
    else:
        pj = evaluate_Nx(N, logx, area_list)
    return pj


# TODO lift to runcard?
_atol_eps = 10 * np.finfo(float).eps


@nb.njit(cache=True)
def evaluate_x(x, area_list):
    """
    Get a single linear Lagrange interpolator in x-space

    .. math::
        p(x)

    Parameters
    ----------
        x : float
            interpolation point
        area_list : list
            area configuration of basis function

    Returns
    -------
        res : float
            basis function(x)
    """
    res = 0.0
    for j, a in enumerate(area_list):
        xmin = a[0]
        xmax = a[1]
        coefs = a[2:]
        if xmin < x <= xmax or (j == 0 and np.abs(x - xmin) < _atol_eps):
            for i, coef in enumerate(coefs):
                res += coef * pow(x, i)
            return res

    return res


@nb.njit(cache=True)
def log_evaluate_x(x, area_list):
    """
    Get a single logarithmic Lagrange interpolator in x-space

    .. math::
        p(x)

    Parameters
    ----------
        x : float
            interpolation point
        area_list : list
            area configuration of basis function

    Returns
    -------
        res : float
            basis function(x)
    """
    x = np.log(x)
    return evaluate_x(x, area_list)


class BasisFunction:
    """
    Object containing a list of areas for a given polynomial number
    defined by (xmin-xmax) and containing a list of coefficients.

    Upon construction will generate all areas and generate and compile
    a function to evaluate in N (or x) the interpolator

    Parameters
    ----------
        xgrid : numpy.ndarray
            Grid in x-space from which the interpolators are constructed
        poly_number : int
            number of polynomial
        list_of_blocks: list(tuple(int, int))
            list of tuples with the (kmin, kmax) values for each area
        mode_log: bool
            use logarithmic interpolation?
        mode_N: bool
            if true compiles the function on N, otherwise compiles x
    """

    def __init__(
        self,
        xgrid,
        poly_number,
        list_of_blocks,
        mode_log=True,
        mode_N=True,
    ):
        self.poly_number = poly_number
        self.areas = []
        self._mode_log = mode_log
        self.mode_N = mode_N

        # create areas
        for i, block in enumerate(list_of_blocks):
            if block[0] <= poly_number <= block[1]:
                new_area = Area(i, self.poly_number, block, xgrid)
                self.areas.append(new_area)
        if not self.areas:
            raise ValueError("Error: no areas were generated")
        self.areas_representation = self.areas_to_const()

        # compile
        # TODO move this to InterpolatorDispatcher
        self.callable = None
        if self.mode_N:
            self.compile_n()
        else:
            self.compile_x()

    def is_below_x(self, x):
        """
        Are all areas below x?

        Parameters
        ----------
            x : float
                reference value
        Returns
        --------
            is_below_x : bool
                xmax of highest area <= x?
        """
        # Log if needed
        if self._mode_log:
            x = np.log(x)
        # note that ordering is important!
        return self.areas[-1].xmax <= x

    def areas_to_const(self):
        """
        Returns an array containing all areas with
        (`xmin`, `xmax`, `numpy.array` of coefficients)

        Returns
        -------
            numpy.ndarray
                area config
        """
        # This is necessary as numba will ask for everything
        # to be immutable
        area_list = []
        for area in self:
            area_list.append([area.xmin, area.xmax, *area.coefs])
        return np.array(area_list)

    def compile_x(self):
        """
        Compiles the function to evaluate the interpolator
        in x space
        """
        if self._mode_log:
            self.callable = log_evaluate_x
        else:
            self.callable = evaluate_x

    def evaluate_x(self, x):
        """
        Evaluate basis function in x-space (regardless of the true space).

        Parameters
        ----------
            x : float
                evaluated point

        Returns
        -------
            res : float
                p(x)
        """
        if self.mode_N:
            old_call = self.callable
            self.compile_x()
            res = self.callable(x, self.areas_representation)
            self.callable = old_call
        else:
            res = self.callable(x, self.areas_representation)
        return res

    def compile_n(self):
        r"""
        Compiles the function to evaluate the interpolator in N space.

        Generates a function :meth:`evaluate_Nx` with a (N, logx) signature.

        .. math::
            \tilde p(N)*\exp(- N * \ln(x))

        The polynomials contain naturally factors of :math:`\exp(N * j * \ln(x_{min/max}))`
        which can be joined with the Mellin inversion factor.
        """

        if self._mode_log:
            self.callable = log_evaluate_Nx
        else:
            self.callable = evaluate_Nx

    def __iter__(self):
        for area in self.areas:
            yield area

    def __call__(self, *args, **kwargs):
        args = list(args)
        args.append(self.areas_representation)
        return self.callable(*args, **kwargs)


class InterpolatorDispatcher:
    """
    Setups the interpolator.

    Upon construction will generate a list of :class:`BasisFunction` objects.
    Each of these :class:`BasisFunction` objects expose a `callable`
    method (also accessible as the `__call__` method of the class)
    which will be numba-compiled.


    Parameters
    ----------
        xgrid_in : numpy.ndarray
            Grid in x-space from which the interpolators are constructed
        polynomial_degree : int
            degree of the interpolation polynomial
        log: bool
            Whether it is a log or linear interpolator
        mode_N: bool
            if true compiles the function on N, otherwise compiles x
    """

    def __init__(self, xgrid, polynomial_degree, log=True, mode_N=True):
        # sanity checks
        xgrid_size = len(xgrid)
        ugrid = np.array(np.unique(xgrid), np.float_)
        if xgrid_size != len(ugrid):
            raise ValueError(f"xgrid is not unique: {xgrid}")
        xgrid = ugrid
        if xgrid_size < 2:
            raise ValueError(f"xgrid needs at least 2 points, received {xgrid_size}")
        if polynomial_degree < 1:
            raise ValueError(
                f"need at least polynomial_degree 1, received {polynomial_degree}"
            )
        if xgrid_size <= polynomial_degree:
            raise ValueError(
                f"to interpolate with degree {polynomial_degree} "
                " we need at least that much points + 1"
            )
        # keep a true copy of grid
        self.xgrid_raw = xgrid
        # henceforth xgrid might no longer be the input!
        # which is ok, because for most of the code this is all we need to do
        # to distinguish log and non-log
        if log:
            xgrid = np.log(xgrid)

        # Save the different variables
        self.xgrid = xgrid
        self.polynomial_degree = polynomial_degree
        self.log = log
        logger.info(
            "Interpolation: number of points = %d, polynomial degree = %d, logarithmic = %s",
            xgrid_size,
            polynomial_degree,
            log,
        )

        # Create blocks
        list_of_blocks = []
        po2 = polynomial_degree // 2
        # if degree is even, we can not split the block symmetric, e.g. deg=2 -> |-|-|
        # so, in case of doubt use the block, which lays higher, i.e.
        # we're not allowed to go so deep -> make po2 smaller
        if polynomial_degree % 2 == 0:
            po2 -= 1
        # iterate areas: there is 1 less then number of points
        for i in range(xgrid_size - 1):
            kmin = max(0, i - po2)
            kmax = kmin + polynomial_degree
            if kmax >= xgrid_size:
                kmax = xgrid_size - 1
                kmin = kmax - polynomial_degree
            b = (kmin, kmax)
            list_of_blocks.append(b)

        # Generate the basis functions
        basis_functions = []
        for i in range(xgrid_size):
            new_basis = BasisFunction(
                xgrid, i, list_of_blocks, mode_log=log, mode_N=mode_N
            )
            basis_functions.append(new_basis)
        self.basis = basis_functions

    @classmethod
    def from_dict(cls, operators_card, mode_N=True):
        """
        Create object from dictionary.

        .. list-table:: operators runcard parameters
            :header-rows: 1

            *   - Name
                - Type
                - description
            *   - ``interpolation_xgrid``
                - :py:obj:`list(float)`
                - the interpolation grid
            *   - ``interpolation_polynomial_degree``
                - :py:obj:`int`
                - polynomial degree of the interpolating function
            *   - ``interpolation_is_log``
                - :py:obj:`bool`
                - use logarithmic interpolation?

        Parameters
        ----------
            operators_card : dict
                input configurations
        """
        # load xgrid
        xgrid = operators_card["interpolation_xgrid"]
        if len(xgrid) == 0:
            raise ValueError("Empty xgrid!")
        if xgrid[0] == "make_grid":
            xgrid = make_grid(*xgrid[1:])
        elif xgrid[0] == "make_lambert_grid":
            xgrid = make_lambert_grid(*xgrid[1:])
        is_log_interpolation = bool(operators_card["interpolation_is_log"])
        polynom_rank = operators_card["interpolation_polynomial_degree"]
        return cls(
            xgrid,
            polynom_rank,
            log=is_log_interpolation,
            mode_N=mode_N,
        )

    def __eq__(self, other):
        """Checks equality"""
        checks = [
            len(self.xgrid_raw) == len(other.xgrid_raw),
            self.log == other.log,
            self.polynomial_degree == other.polynomial_degree,
        ]
        # check elements after shape
        return all(checks) and np.allclose(self.xgrid_raw, other.xgrid_raw)

    def __iter__(self):
        # return iter(self.basis)
        for basis in self.basis:
            yield basis

    def __getitem__(self, item):
        return self.basis[item]

    def get_interpolation(self, targetgrid):
        """Computes interpolation matrix between `targetgrid` and `xgrid`.

        .. math::
            f(targetgrid) = R \\cdot f(xgrid)

        Parameters
        ----------
        targetgrid : array
            grid to interpolate to

        Returns
        -------
        R : array
            interpolation matrix $R_{ij}$, where $i$ is the index over
            `targetgrid`, and $j$ is the index on the internal basis (the
            one stored in the :class:`InterpolatorDispatcher` instance)

        """
        # trivial?
        if len(targetgrid) == len(self.xgrid_raw) and np.allclose(
            targetgrid, self.xgrid_raw
        ):
            return np.eye(len(self.xgrid_raw))
        # compute map
        out = []
        for x in targetgrid:
            row = []
            for b in self.basis:
                row.append(b.evaluate_x(x))
            out.append(row)
        return np.array(out)

    def to_dict(self):
        """
        Returns the configuration for the underlying xgrid (from which the instance can
        be created again).

        The output dictionary contains a numpy array.

        Returns
        -------
            ret : dict
                full grid configuration
        """
        ret = {
            "interpolation_xgrid": self.xgrid_raw,
            "interpolation_polynomial_degree": self.polynomial_degree,
            "interpolation_is_log": self.log,
        }
        return ret


def make_grid(
    n_low, n_mid, n_high=0, x_min=1e-7, x_low=0.1, x_high=0.9, x_high_max=1.0 - 1e-4
):
    """
    Creates a log-lin-log-spaced grid.

    1.0 is always part of the grid and the final grid is unified, i.e. esp. points that might
    appear twice at the borders of the regions are unified.

    Parameters
    ----------
        n_low : int
            points in the small-x region
        n_mid : int
            points in the medium-x region
        n_high : int
            points in the large-x region
        x_min : float
            minimum x (included)
        x_low : float
            seperation point between small and medium
        x_high_max : float
            closest point before 1

    Returns
    -------
        xgrid : numpy.ndarray
            generated grid
    """
    # low
    if n_mid == 0:
        x_low = 1.0
    xgrid_low = np.geomspace(x_min, x_low, n_low)
    # high
    if n_high == 0:
        x_high = 1.0
        xgrid_high = np.array([])
    else:
        xgrid_high = 1.0 - np.geomspace(1 - x_high, 1 - x_high_max, n_high)
    # mid
    xgrid_mid = np.linspace(x_low, x_high, n_mid)
    # join
    xgrid = np.unique(np.concatenate((xgrid_low, xgrid_mid, xgrid_high, np.array([1]))))
    return xgrid


def make_lambert_grid(
    n_pts,
    x_min=1e-7,
    x_max=1.0,
):
    r"""
    Creates a smoothly spaced grid that is linear near 1 and logarithmic near 0.

    It is generated by the relation:

    .. math::
        x(y) = \frac{1}{5} W_{0}(5 \exp(5-y))

    where :math:`W_{0}` is the principle branch of the
    :func:`Lambert W function <scipy.special.lambertw>` and
    :math:`y` is a variable which extremes are given as function of :math:`x`
    by the direct relation:

    .. math::
        y(x) = 5(1-x)-\log(x)

    This method is implemented in `PineAPPL`, :cite:`Carrazza_2020` eq 2.11 and relative
    paragraph.

    Parameters
    ----------
        n_pts : int
            number of points
        x_min : float
            minimum x (included)
        x_max : float
            maximum x (included)

    Returns
    -------
        xgrid : numpy.ndarray
            generated grid
    """

    def direct_relation(x):
        return 5 * (1 - x) - np.log(x)

    def inverse_relation(y):
        return np.real(1 / 5 * sp.lambertw(5 * np.exp(5 - y)))

    y_min = direct_relation(x_min)
    y_max = direct_relation(x_max)

    return np.array([inverse_relation(y) for y in np.linspace(y_min, y_max, n_pts)])



# -*- coding: utf-8 -*-
r"""Holds the classes that define the |FNS|."""
import logging

import numpy as np

logger = logging.getLogger(__name__)


class PathSegment:
    """Oriented path in the threshold landscape.

    Attributes
    ----------
    q2_from : float
        starting point
    q2_to : float
        final point
    nf : int
        number of active flavors

    Parameters
    ----------
    q2_from : float
        starting point
    q2_to : float
        final point
    nf : int
        number of active flavors
    """

    def __init__(self, q2_from, q2_to, nf):
        self.q2_from = q2_from
        self.q2_to = q2_to
        self.nf = nf

    @property
    def is_downward_q2(self):
        """Return True if ``q2_from`` is bigger than ``q2_to``."""
        return self.q2_from > self.q2_to

    @property
    def tuple(self):
        """Return tuple representation suitable for hashing."""
        return (self.q2_from, self.q2_to, self.nf)

    def __repr__(self):
        """Return string representation."""
        return f"PathSegment({self.q2_from} -> {self.q2_to}, nf={self.nf})"


class ThresholdsAtlas:
    r"""Holds information about the matching scales that any :math:`Q^2` has to pass in order to get there from a given :math:`Q^2_{ref}`."""

    def __init__(
        self,
        masses,
        q2_ref=None,
        nf_ref=None,
        thresholds_ratios=None,
        max_nf=None,
    ):
        """Create basic atlas.

        Parameters
        ----------
        masses : list(float)
            list of quark masses squared
        q2_ref : float
            reference scale
        nf_ref : int
            number of active flavors at the reference scale
        thresholds_ratios : list(float)
            list of ratios between masses and matching scales squared
        max_nf : int
            maximum number of active flavors
        """
        masses = list(masses)
        if masses != sorted(masses):
            raise ValueError("masses need to be sorted")
        # combine them
        thresholds = self.build_area_walls(masses, thresholds_ratios, max_nf)
        self.area_walls = [0] + thresholds + [np.inf]

        # check nf_ref
        if nf_ref is not None:
            if q2_ref is None:
                raise ValueError(
                    "Without a reference Q2 value a reference number of flavors "
                    "does not make sense!"
                )
            # else self.q2_ref is not None
            nf_init = 2 + len(list(filter(lambda x: np.isclose(0, x), self.area_walls)))
            if nf_ref < nf_init:
                raise ValueError(
                    f"The reference number of flavors is set to {nf_ref}, "
                    f"but the atlas starts at {nf_init}"
                )
            nf_final = 2 + len(list(filter(lambda x: x < np.inf, self.area_walls)))
            if nf_ref > nf_final:
                raise ValueError(
                    f"The reference number of flavors is set to {nf_ref}, "
                    f"but the atlas stops at {nf_final}"
                )

        # Init values
        self.q2_ref = q2_ref
        self.nf_ref = nf_ref
        self.thresholds_ratios = thresholds_ratios
        logger.info(str(self))

    def __repr__(self):
        """Return string representation."""
        walls = " - ".join([f"{w:.2e}" for w in self.area_walls])
        return f"ThresholdsAtlas [{walls}], ref={self.q2_ref} @ {self.nf_ref}"

    @classmethod
    def ffns(cls, nf, q2_ref=None):
        """Create a |FFNS| setup.

        The function creates simply sufficient thresholds at ``0`` (in the
        beginning), since the number of flavors is determined by counting
        from below.

        Parameters
        ----------
        nf : int
            number of light flavors
        q2_ref : float
            reference scale
        """
        return cls([0] * (nf - 3) + [np.inf] * (6 - nf), q2_ref)

    @staticmethod
    def build_area_walls(masses, thresholds_ratios=None, max_nf=None):
        r"""Create the object from the informations on the run card.

        The thresholds are computed by :math:`(m_q \cdot k_q^{Thr})`.

        Parameters
        ----------
        masses : list(float)
            heavy quark masses squared
        thresholds_ratios : list(float)
            list of ratios between masses and matching scales squared
        max_nf : int
            maximum number of flavors

        Returns
        -------
        list
            threshold list
        """
        if len(masses) != 3:
            raise ValueError("There have to be 3 quark masses")
        if thresholds_ratios is None:
            thresholds_ratios = (1.0, 1.0, 1.0)
        if len(thresholds_ratios) != 3:
            raise ValueError("There have to be 3 quark threshold ratios")
        if max_nf is None:
            max_nf = 6

        thresholds = []
        for m, k in zip(masses, thresholds_ratios):
            thresholds.append(m * k)
        # cut array = simply reduce some thresholds
        thresholds = thresholds[: max_nf - 3]
        return thresholds

    @classmethod
    def from_dict(cls, theory_card, prefix="k", max_nf_name="MaxNfPdf", masses=None):
        r"""Create the atlas from the run card.

        The thresholds are computed by :math:`(m_q \cdot k_q^{Thr})`.

        Parameters
        ----------
        theory_card : dict
            run card with the keys given at the head of the :mod:`module <eko.thresholds>`
        prefix : str
            prefix for the ratio parameters
        masses : list
            list of |MSbar| masses squared or None if POLE masses are used

        Returns
        -------
        ThresholdsAtlas
            created atlas
        """
        heavy_flavors = "cbt"
        if masses is None:
            masses = np.power([theory_card[f"m{q}"] for q in heavy_flavors], 2)
        thresholds_ratios = [theory_card[f"{prefix}{q}Thr"] for q in heavy_flavors]
        max_nf = theory_card[max_nf_name]
        # preset ref scale
        q2_ref = pow(theory_card["Q0"], 2)

        # MSbar or Pole masses
        hqm_scheme = theory_card["HQ"]
        if hqm_scheme not in ["MSBAR", "POLE"]:
            raise ValueError(f"{hqm_scheme} is not implemented, choose POLE or MSBAR")
        nf_ref = theory_card["nf0"]
        return cls(
            masses,
            q2_ref,
            nf_ref,
            thresholds_ratios=np.power(thresholds_ratios, 2),
            max_nf=max_nf,
        )

    def path(self, q2_to, nf_to=None, q2_from=None, nf_from=None):
        """Get path from ``q2_from`` to ``q2_to``.

        Parameters
        ----------
        q2_to: float
            target value of q2
        q2_from: float
            starting value of q2

        Returns
        -------
        list(PathSegment)
            List of :class:`PathSegment` to go through in order to get from ``q2_from``
            to ``q2_to``.
        """
        # fallback to init config
        if q2_from is None:
            q2_from = self.q2_ref
        if nf_from is None:
            nf_from = self.nf_ref
        # determine reference thresholds
        if nf_from is None:
            nf_from = 2 + np.digitize(q2_from, self.area_walls)
        if nf_to is None:
            nf_to = 2 + np.digitize(q2_to, self.area_walls)
        # determine direction and python slice modifier
        if nf_to < nf_from:
            rc = -1
            shift = -3
        else:
            rc = 1
            shift = -2
        # join all necessary points in one list
        boundaries = (
            [q2_from]
            + self.area_walls[nf_from + shift : int(nf_to) + shift : rc]
            + [q2_to]
        )
        segs = [
            PathSegment(boundaries[i], q2, nf_from + i * rc)
            for i, q2 in enumerate(boundaries[1:])
        ]
        return segs

    def nf(self, q2):
        """Find the number of flavors active at the given scale.

        Parameters
        ----------
        q2 : float
            reference scale

        Returns
        -------
        int
            number of active flavors
        """
        ref_idx = np.digitize(q2, self.area_walls)
        return 2 + ref_idx


def is_downward_path(path):
    """Determine if a path is downward.

    Criterias are:

    - in the number of active flavors when the path list contains more than one :class:`PathSegment`,
      note this can be different from each :attr:`PathSegment.is_downward_q2`
    - in :math:`Q^2` when just one single :class:`PathSegment` is given

    Parameters
    ----------
    path : list(PathSegment)
        path

    Returns
    -------
    bool
        True for a downward path
    """
    if len(path) == 1:
        return path[0].is_downward_q2
    return path[1].nf < path[0].nf


def flavor_shift(is_downward):
    """Determine the shift to number of light flavors.

    Parameters
    ----------
    is_downward : bool
        True for a downward path

    Returns
    -------
    int
        shift to number of light flavors which can be 3 or 4
    """
    return 4 if is_downward else 3



# -*- coding: utf-8 -*-
"""
Collection of singlet EKOs.
"""

import numba as nb
import numpy as np

from .. import anomalous_dimensions as ad
from .. import beta
from . import as4_evolution_integrals as as4_ei
from . import evolution_integrals as ei
from . import utils


@nb.njit(cache=True)
def lo_exact(gamma_singlet, a1, a0, nf):
    """
    Singlet leading order exact EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_s^0 : numpy.ndarray
            singlet leading order exact EKO
    """
    return ad.exp_singlet(gamma_singlet[0] * ei.j00(a1, a0, nf))[0]


@nb.njit(cache=True)
def nlo_decompose(gamma_singlet, j01, j11):
    """
    Singlet next-to-leading order decompose EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors
        j01 : float
            LO-NLO evolution integral
        j11 : float
            NLO-NLO evolution integral

    Returns
    -------
        e_s^1 : numpy.ndarray
            singlet next-to-leading order decompose EKO
    """
    return ad.exp_singlet(gamma_singlet[0] * j01 + gamma_singlet[1] * j11)[0]


@nb.njit(cache=True)
def nlo_decompose_exact(gamma_singlet, a1, a0, nf):
    """
    Singlet next-to-leading order decompose-exact EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_s^1 : numpy.ndarray
            singlet next-to-leading order decompose-exact EKO
    """
    return nlo_decompose(
        gamma_singlet, ei.j01_exact(a1, a0, nf), ei.j11_exact(a1, a0, nf)
    )


@nb.njit(cache=True)
def nlo_decompose_expanded(gamma_singlet, a1, a0, nf):
    """
    Singlet next-to-leading order decompose-expanded EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_s^1 : numpy.ndarray
            singlet next-to-leading order decompose-expanded EKO
    """
    return nlo_decompose(
        gamma_singlet, ei.j01_expanded(a1, a0, nf), ei.j11_expanded(a1, a0, nf)
    )


@nb.njit(cache=True)
def nnlo_decompose(gamma_singlet, j02, j12, j22):
    """
    Singlet next-to-next-to-leading order decompose EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors
        j02 : float
            LO-NNLO evolution integral
        j12 : float
            NLO-NNLO evolution integral
        j22 : float
            NNLO-NNLO evolution integral
    Returns
    -------
        e_s^2 : numpy.ndarray
            singlet next-to-next-to-leading order decompose EKO
    """
    return ad.exp_singlet(
        gamma_singlet[0] * j02 + gamma_singlet[1] * j12 + gamma_singlet[2] * j22
    )[0]


@nb.njit(cache=True)
def nnlo_decompose_exact(gamma_singlet, a1, a0, nf):
    """
    Singlet next-to-next-to-leading order decompose-exact EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_s^2 : numpy.ndarray
            singlet next-to-next-to-leading order decompose-exact EKO
    """
    return nnlo_decompose(
        gamma_singlet,
        ei.j02_exact(a1, a0, nf),
        ei.j12_exact(a1, a0, nf),
        ei.j22_exact(a1, a0, nf),
    )


@nb.njit(cache=True)
def nnlo_decompose_expanded(gamma_singlet, a1, a0, nf):
    """
    Singlet next-to-next-to-leading order decompose-expanded EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_s^2 : numpy.ndarray
            singlet next-to-next-to-leading order decompose-expanded EKO
    """
    return nnlo_decompose(
        gamma_singlet,
        ei.j02_expanded(a1, a0, nf),
        ei.j12_expanded(a1, a0, nf),
        ei.j22_expanded(a1, a0, nf),
    )


@nb.njit(cache=True)
def n3lo_decompose(gamma_singlet, j03, j13, j23, j33):
    """Singlet |N3LO| decompose EKO

    Parameters
    ----------
    gamma_singlet : numpy.ndarray
        singlet anomalous dimensions matrices
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    nf : int
        number of active flavors
    j03 : float
        |LO|-|N3LO| evolution integral
    j13 : float
        |NLO|-|N3LO| evolution integral
    j23 : float
        |NNLO|-|N3LO| evolution integral
    j33 : float
        |N3LO|-|N3LO| evolution integral

    Returns
    -------
    numpy.ndarray
        singlet |N3LO| decompose EKO

    """
    return ad.exp_singlet(
        gamma_singlet[0] * j03
        + gamma_singlet[1] * j13
        + gamma_singlet[2] * j23
        + gamma_singlet[3] * j33
    )[0]


@nb.njit(cache=True)
def n3lo_decompose_exact(gamma_singlet, a1, a0, nf):
    """Singlet |N3LO| decompose-exact EKO

    Parameters
    ----------
    gamma_singlet : numpy.ndarray
        singlet anomalous dimensions matrices
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    nf : int
        number of active flavors

    Returns
    -------
    numpy.ndarray
        singlet |N3LO| decompose-exact EKO

    """
    beta0 = beta.beta_qcd((2, 0), nf)
    b_list = [
        beta.b_qcd((3, 0), nf),
        beta.b_qcd((4, 0), nf),
        beta.b_qcd((5, 0), nf),
    ]
    roots = as4_ei.roots(b_list)
    j00 = ei.j00(a1, a0, nf)
    j13 = as4_ei.j13_exact(a1, a0, beta0, b_list, roots)
    j23 = as4_ei.j23_exact(a1, a0, beta0, b_list, roots)
    j33 = as4_ei.j33_exact(a1, a0, beta0, b_list, roots)
    return n3lo_decompose(
        gamma_singlet, as4_ei.j03_exact(j00, j13, j23, j33, b_list), j13, j23, j33
    )


@nb.njit(cache=True)
def n3lo_decompose_expanded(gamma_singlet, a1, a0, nf):
    """Singlet |N3LO| decompose-expanded EKO

    Parameters
    ----------
    gamma_singlet : numpy.ndarray
        singlet anomalous dimensions matrices
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    nf : int
        number of active flavors

    Returns
    -------
    numpy.ndarray
        singlet |N3LO| decompose-expanded EKO

    """
    beta0 = beta.beta_qcd((2, 0), nf)
    b_list = [
        beta.b_qcd((3, 0), nf),
        beta.b_qcd((4, 0), nf),
        beta.b_qcd((5, 0), nf),
    ]
    j00 = ei.j00(a1, a0, nf)
    j13 = as4_ei.j13_expanded(a1, a0, beta0, b_list)
    j23 = as4_ei.j23_expanded(a1, a0, beta0, b_list)
    j33 = as4_ei.j33_expanded(a1, a0, beta0)
    return n3lo_decompose(
        gamma_singlet, as4_ei.j03_expanded(j00, j13, j23, j33, b_list), j13, j23, j33
    )


@nb.njit(cache=True)
def eko_iterate(gamma_singlet, a1, a0, nf, order, ev_op_iterations):
    """Singlet |NLO|, |NNLO| or |N3LO| iterated (exact) EKO

    Parameters
    ----------
    gamma_singlet : numpy.ndarray
        singlet anomalous dimensions matrices
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    nf : int
        number of active flavors
    order : tuple(int,int)
        perturbative order
    ev_op_iterations : int
        number of evolution steps

    Returns
    -------
    numpy.ndarray
        singlet iterated (exact) EKO

    """
    a_steps = utils.geomspace(a0, a1, 1 + ev_op_iterations)
    beta_vec = [beta.beta_qcd((2, 0), nf), beta.beta_qcd((3, 0), nf)]
    if order[0] >= 3:
        beta_vec.append(beta.beta_qcd((4, 0), nf))
    if order[0] >= 4:
        beta_vec.append(beta.beta_qcd((5, 0), nf))
    e = np.identity(2, np.complex_)
    al = a_steps[0]
    for ah in a_steps[1:]:
        a_half = (ah + al) / 2.0
        delta_a = ah - al
        gamma_summed = np.zeros((2, 2), dtype=np.complex_)
        beta_summed = 0
        for i in range(order[0]):
            gamma_summed += gamma_singlet[i] * a_half**i
            beta_summed += beta_vec[i] * a_half ** (i + 1)
        ln = gamma_summed / beta_summed * delta_a
        ek = np.ascontiguousarray(ad.exp_singlet(ln)[0])
        e = ek @ e
        al = ah
    return e


@nb.njit(cache=True)
def r_vec(gamma_singlet, nf, ev_op_max_order, order, is_exact):
    r"""Compute singlet R vector for perturbative mode.

    .. math::
        \frac{d}{da_s} \dSV{1}{a_s} &= \frac{\mathbf R (a_s)}{a_s} \cdot \dSV{1}{a_s}\\
        \mathbf R (a_s) &= \sum\limits_{k=0} a_s^k \mathbf R_{k}

    Parameters
    ----------
    gamma_singlet : numpy.ndarray
        singlet anomalous dimensions matrices
    nf : int
        number of active flavors
    ev_op_max_order : tuple(int,int)
        perturbative expansion order of U
    order : tuple(int,int)
       perturbative order
    is_exact : boolean
        fill up r-vector?

    Returns
    -------
    np.ndarray
        R vector

    """
    r = np.zeros(
        (ev_op_max_order[0] + 1, 2, 2), dtype=np.complex_
    )  # k = 0 .. max_order
    beta0 = beta.beta_qcd((2, 0), nf)
    # fill explicit elements
    r[0] = gamma_singlet[0] / beta0
    if order[0] > 1:
        b1 = beta.b_qcd((3, 0), nf)
        r[1] = gamma_singlet[1] / beta0 - b1 * r[0]
    if order[0] > 2:
        b2 = beta.b_qcd((4, 0), nf)
        r[2] = gamma_singlet[2] / beta0 - b1 * r[1] - b2 * r[0]
    if order[0] > 3:
        b3 = beta.b_qcd((5, 0), nf)
        r[3] = gamma_singlet[3] / beta0 - b1 * r[2] - b2 * r[1] - b3 * r[0]
    # fill rest
    if is_exact:
        if order[0] == 2:
            for kk in range(2, ev_op_max_order[0]):
                r[kk] = -b1 * r[kk - 1]
        elif order[0] == 3:
            for kk in range(3, ev_op_max_order[0]):
                r[kk] = -b1 * r[kk - 1] - b2 * r[kk - 2]
        elif order[0] == 4:
            for kk in range(4, ev_op_max_order[0] + 1):
                r[kk] = -b1 * r[kk - 1] - b2 * r[kk - 2] - b3 * r[kk - 3]
    return r


@nb.njit(cache=True)
def u_vec(r, ev_op_max_order):
    r"""Compute the elements of the singlet U vector.

    .. math::
        \ESk{n}{a_s}{a_s^0} &= \mathbf U (a_s) \ESk{0}{a_s}{a_s^0} {\mathbf U}^{-1} (a_s^0)\\
        \mathbf U (a_s) &= \mathbf I + \sum\limits_{k=1} a_s^k \mathbf U_k

    Parameters
    ----------
    r : numpy.ndarray
        singlet R vector
    ev_op_max_order : tuple(int,int)
        perturbative expansion order of U

    Returns
    -------
    numpy.ndarray
        U vector

    """
    u = np.zeros((ev_op_max_order[0], 2, 2), np.complex_)  # k = 0 .. max_order
    # init
    u[0] = np.identity(2, np.complex_)
    _, r_p, r_m, e_p, e_m = ad.exp_singlet(r[0])
    e_p = np.ascontiguousarray(e_p)
    e_m = np.ascontiguousarray(e_m)
    for kk in range(1, ev_op_max_order[0]):
        # compute R'
        rp = np.zeros((2, 2), dtype=np.complex_)
        for jj in range(kk):
            rp += np.ascontiguousarray(r[kk - jj]) @ u[jj]
        # now compose U
        u[kk] = (
            (e_m @ rp @ e_m + e_p @ rp @ e_p) / kk
            + ((e_p @ rp @ e_m) / (r_m - r_p + kk))
            + ((e_m @ rp @ e_p) / (r_p - r_m + kk))
        )
    return u


@nb.njit(cache=True)
def sum_u(uvec, a):
    r"""Sums up the actual U operator.


    .. math::
        \mathbf U (a_s) = \mathbf I + \sum\limits_{k=1} a_s^k \mathbf U_k

    Parameters
    ----------
    uvec : numpy.ndarray
        U vector (elements)
    a : float
        strong coupling

    Returns
    -------
    numpy.ndarray
        sum

    """
    p = 1.0
    res = np.zeros((2, 2), dtype=np.complex_)
    for uk in uvec:
        res += p * uk
        p *= a
    # alternative implementation:
    # al_vec = al**(np.arange(len(uk)))
    # ul = np.sum(al_vec * uk.T,-1).T
    return res


@nb.njit(cache=True)
def eko_perturbative(
    gamma_singlet, a1, a0, nf, order, ev_op_iterations, ev_op_max_order, is_exact
):
    """Singlet |NLO|,|NNLO| or |N3LO| perturbative EKO, depending on which r is passed

    Parameters
    ----------
    gamma_singlet : numpy.ndarray
        singlet anomalous dimensions matrices
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    nf : int
        number of active flavors
    order : tuple(int,int)
        perturbative order
    ev_op_iterations : int
        number of evolution steps
    ev_op_max_order : tuple(int,int)
        perturbative expansion order of U
    is_exact : boolean
        fill up r-vector?

    Returns
    -------
    numpy.ndarray
        singlet perturbative EKO

    """
    r = r_vec(gamma_singlet, nf, ev_op_max_order, order, is_exact)
    uk = u_vec(r, ev_op_max_order)
    e = np.identity(2, np.complex_)
    # iterate elements
    a_steps = utils.geomspace(a0, a1, 1 + ev_op_iterations)
    al = a_steps[0]
    for ah in a_steps[1:]:
        e0 = lo_exact(gamma_singlet, ah, al, nf)
        uh = sum_u(uk, ah)
        ul = sum_u(uk, al)
        # join elements
        ek = np.ascontiguousarray(uh) @ np.ascontiguousarray(e0) @ np.linalg.inv(ul)
        e = ek @ e
        al = ah
    return e


@nb.njit(cache=True)
def eko_truncated(gamma_singlet, a1, a0, nf, order, ev_op_iterations):
    """Singlet |NLO|, |NNLO| or |N3LO| truncated EKO

    Parameters
    ----------
    gamma_singlet : numpy.ndarray
        singlet anomalous dimensions matrices
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    nf : int
        number of active flavors
    order : tuple(int,int)
        perturbative order
    ev_op_iterations : int
        number of evolution steps

    Returns
    -------
    numpy.ndarray
        singlet truncated EKO

    """
    r = r_vec(gamma_singlet, nf, order, order, False)
    u = u_vec(r, order)
    u1 = np.ascontiguousarray(u[1])
    e = np.identity(2, np.complex_)
    # iterate elements
    a_steps = utils.geomspace(a0, a1, 1 + ev_op_iterations)
    al = a_steps[0]
    for ah in a_steps[1:]:
        e0 = np.ascontiguousarray(lo_exact(gamma_singlet, ah, al, nf))
        if order[0] >= 2:
            ek = e0 + ah * u1 @ e0 - al * e0 @ u1
        if order[0] >= 3:
            u2 = np.ascontiguousarray(u[2])
            ek += (
                +(ah**2) * u2 @ e0
                - ah * al * u1 @ e0 @ u1
                + al**2 * e0 @ (u1 @ u1 - u2)
            )
        if order[0] >= 4:
            u3 = np.ascontiguousarray(u[3])
            ek += (
                +(ah**3) * u3 @ e0
                - ah**2 * al * u2 @ e0 @ u1
                + ah * al**2 * u1 @ e0 @ (u1 @ u1 - u2)
                - al**3 * e0 @ (u1 @ u1 @ u1 - u1 @ u2 - u2 @ u1 + u3)
            )
        e = ek @ e
        al = ah
    return e


@nb.njit(cache=True)
def dispatcher(  # pylint: disable=too-many-return-statements
    order, method, gamma_singlet, a1, a0, nf, ev_op_iterations, ev_op_max_order
):
    """Determine used kernel and call it.

    In LO we always use the exact solution.

    Parameters
    ----------
    order :  tuple(int,int)
        perturbative order
    method : str
        method
    gamma_singlet : numpy.ndarray
        singlet anomalous dimensions matrices
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    nf : int
        number of active flavors
    ev_op_iterations : int
        number of evolution steps
    ev_op_max_order : tuple(int,int)
        perturbative expansion order of U

    Returns
    -------
    numpy.ndarray
        singlet EKO

    """
    # use always exact in LO
    if order[0] == 1:
        return lo_exact(gamma_singlet, a1, a0, nf)

    # Common method for NLO and NNLO
    if method in ["iterate-exact", "iterate-expanded"]:
        return eko_iterate(gamma_singlet, a1, a0, nf, order, ev_op_iterations)
    if method == "perturbative-exact":
        return eko_perturbative(
            gamma_singlet, a1, a0, nf, order, ev_op_iterations, ev_op_max_order, True
        )
    if method == "perturbative-expanded":
        return eko_perturbative(
            gamma_singlet, a1, a0, nf, order, ev_op_iterations, ev_op_max_order, False
        )
    if method in ["truncated", "ordered-truncated"]:
        return eko_truncated(gamma_singlet, a1, a0, nf, order, ev_op_iterations)
    # These methods are scattered for nlo and nnlo
    if method == "decompose-exact":
        if order[0] == 2:
            return nlo_decompose_exact(gamma_singlet, a1, a0, nf)
        elif order[0] == 3:
            return nnlo_decompose_exact(gamma_singlet, a1, a0, nf)
        return n3lo_decompose_exact(gamma_singlet, a1, a0, nf)
    if method == "decompose-expanded":
        if order[0] == 2:
            return nlo_decompose_expanded(gamma_singlet, a1, a0, nf)
        elif order[0] == 3:
            return nnlo_decompose_expanded(gamma_singlet, a1, a0, nf)
        return n3lo_decompose_expanded(gamma_singlet, a1, a0, nf)
    raise NotImplementedError("Selected method is not implemented")



# -*- coding: utf-8 -*-
import copy

import numpy as np


# TODO: add a control on the theory (but before we need to implement another
# kind of output which includes the theory and operator runcards)
def ekos_product(eko_ini, eko_fin, in_place=True):
    """Returns the product of two ekos

    Parameters
    ----------
    eko_ini : eko.output.Output
        initial eko operator
    eko_fin : eko.output.Output
        final eko operator
    in_place : bool
        do operation in place, modifying input arrays

    Returns
    -------
    eko.output.Output
        eko operator

    """
    if eko_fin["q2_ref"] not in eko_ini["Q2grid"].keys():
        raise ValueError(
            "Initial Q2 of final eko operator does not match any final Q2 in"
            " the initial eko operator"
        )
    ope1 = eko_ini["Q2grid"][eko_fin["q2_ref"]]["operators"]
    ope1_error = eko_ini["Q2grid"][eko_fin["q2_ref"]]["operator_errors"]

    ope2_dict = {}
    ope2_error_dict = {}
    for q2, op in eko_fin["Q2grid"].items():
        ope2_dict[q2] = op["operators"]
        ope2_error_dict[q2] = op["operator_errors"]

    final_op_dict = {}
    final_op_error_dict = {}
    final_dict = {}
    for q2, op2 in ope2_dict.items():
        final_op_dict[q2] = np.einsum("ajbk,bkcl -> ajcl", ope1, op2)

        final_op_error_dict[q2] = np.einsum(
            "ajbk,bkcl -> ajcl", ope1, ope2_error_dict[q2]
        ) + np.einsum("ajbk,bkcl -> ajcl", ope1_error, op2)

        final_dict[q2] = {
            "operators": final_op_dict[q2],
            "operator_errors": final_op_error_dict[q2],
        }

    final_eko = None
    if in_place is False:
        final_eko = copy.deepcopy(eko_ini)
    else:
        final_eko = eko_ini
    final_eko["Q2grid"] = final_dict
    return final_eko



# -*- coding: utf-8 -*-
r"""This module defines the matching conditions for the N3LO |VFNS| evolution.

The expressions are based on:

    - :cite:`Bierenbaum:2009mv`. Isabella Bierenbaum, Johannes Blümlein, and
      Sebastian Klein. Mellin Moments of the O(alpha**3(s)) Heavy Flavor
      Contributions to unpolarized Deep-Inelastic Scattering at Q**2 \ensuremath
      >\ensuremath > m**2 and Anomalous Dimensions. Nucl. Phys. B, 820:417-482,
      2009. arXiv:0904.3563, doi:10.1016/j.nuclphysb.2009.06.005.
    - :cite:`Bl_mlein_2000`. Johannes Blümlein. Analytic continuation of mellin
      transforms up to two-loop order. Computer Physics Communications,
      133(1):76-104, Dec 2000. URL:
      http://dx.doi.org/10.1016/S0010-4655(00)00156-9,
      doi:10.1016/s0010-4655(00)00156-9.
    - :cite:`Bierenbaum:2009zt`. Isabella Bierenbaum, Johannes Blümlein, and
      Sebastian Klein. The Gluonic Operator Matrix Elements at O(alpha(s)**2) for
      DIS Heavy Flavor Production. Phys. Lett. B, 672:401-406, 2009.
      arXiv:0901.0669, doi:10.1016/j.physletb.2009.01.057.
    - :cite:`Ablinger:2010ty`. J. Ablinger, J. Blümlein, S. Klein, C. Schneider,
      and F. Wissbrock. The $O(\alpha _s^3)$ Massive Operator Matrix Elements of
      $O(n_f)$ for the Structure Function $F_2(x,Q^2)$ and Transversity. Nucl.
      Phys. B, 844:26-54, 2011. arXiv:1008.3347,
      doi:10.1016/j.nuclphysb.2010.10.021.
    - :cite:`Ablinger:2014vwa`. J. Ablinger, A. Behring, J. Blümlein, A. De
      Freitas, A. Hasselhuhn, A. von Manteuffel, M. Round, C. Schneider, and F.
      Wißbrock. The 3-Loop Non-Singlet Heavy Flavor Contributions and Anomalous
      Dimensions for the Structure Function $F_2(x,Q^2)$ and Transversity. Nucl.
      Phys. B, 886:733-823, 2014. arXiv:1406.4654,
      doi:10.1016/j.nuclphysb.2014.07.010.
    - :cite:`Ablinger:2014uka`. J. Ablinger, J. Blümlein, A. De Freitas, A.
      Hasselhuhn, A. von Manteuffel, M. Round, and C. Schneider. The $O(\alpha
      _s^3 T_F^2)$ Contributions to the Gluonic Operator Matrix Element. Nucl.
      Phys. B, 885:280-317, 2014. arXiv:1405.4259,
      doi:10.1016/j.nuclphysb.2014.05.028.
    - :cite:`Behring:2014eya`. A. Behring, I. Bierenbaum, J. Blümlein, A. De
      Freitas, S. Klein, and F. Wißbrock. The logarithmic contributions to the
      $O(\alpha ^3_s)$ asymptotic massive Wilson coefficients and operator matrix
      elements in deeply inelastic scattering. Eur. Phys. J. C, 74(9):3033, 2014.
      arXiv:1403.6356, doi:10.1140/epjc/s10052-014-3033-x.
    - :cite:`Blumlein:2017wxd`. Johannes Blümlein, Jakob Ablinger, Arnd Behring,
      Abilio De Freitas, Andreas von Manteuffel, Carsten Schneider, and C.
      Schneider. Heavy Flavor Wilson Coefficients in Deep-Inelastic Scattering:
      Recent Results. PoS, QCDEV2017:031, 2017. arXiv:1711.07957,
      doi:10.22323/1.308.0031.
    - :cite:`Ablinger_2014`. J. Ablinger, J. Blümlein, A. De Freitas, A.
      Hasselhuhn, A. von Manteuffel, M. Round, C. Schneider, and F. Wißbrock. The
      transition matrix element a_gq(n) of the variable flavor number scheme at
      o(α_s^3). Nuclear Physics B, 882:263-288, May 2014. URL:
      http://dx.doi.org/10.1016/j.nuclphysb.2014.02.007,
      doi:10.1016/j.nuclphysb.2014.02.007.
    - :cite:`Ablinger_2015`. J. Ablinger, A. Behring, J. Blümlein, A. De
      Freitas, A. von Manteuffel, and C. Schneider. The 3-loop pure singlet heavy
      flavor contributions to the structure function f2(x,q2) and the anomalous
      dimension. Nuclear Physics B, 890:48-151, Jan 2015. URL:
      http://dx.doi.org/10.1016/j.nuclphysb.2014.10.008,
      doi:10.1016/j.nuclphysb.2014.10.008.
"""

import numba as nb
import numpy as np

from .agg import A_gg
from .agq import A_gq
from .aHg import A_Hg
from .aHq import A_Hq
from .aqg import A_qg
from .aqqNS import A_qqNS
from .aqqPS import A_qqPS


@nb.njit(cache=True)
def A_singlet(n, sx_singlet, sx_non_singlet, nf, L):
    r"""Computes the |N3LO| singlet |OME|.

    .. math::
        A^{S,(3)} = \left(\begin{array}{cc}
            A_{gg, H}^{S,(3)} & A_{gq, H}^{S,(3)} & 0 \\
            A_{qg, H}^{S,(3)} & A_{qq,H}^{NS,(3)} + A_{qq,H}^{PS,(3)} & 0\\
            A_{hg}^{S,(3)} & A_{hq}^{PS,(3)} & 0\\
        \end{array}\right)

    When using the code, please cite the complete list of references
    available at the top of this module :mod:`eko.matching_conditions.as3`.

    Parameters
    ----------
    n : complex
        Mellin moment
    sx_singlet : list
        singlet like harmonic sums cache containing:

        .. math ::
            [[S_1,S_{-1}],
            [S_2,S_{-2}],
            [S_{3}, S_{2,1}, S_{2,-1}, S_{-2,1}, S_{-2,-1}, S_{-3}],
            [S_{4}, S_{3,1}, S_{2,1,1}, S_{-2,-2}, S_{-3, 1}, S_{-4}],]

    sx_non_singlet: list
        same as sx_singlet but now for non-singlet like harmonics
    nf : int
        number of active flavor below the threshold
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
    A_S : numpy.ndarray
        |NNLO| singlet |OME| :math:`A^{S,(3)}(N)`

    """
    A_hq_3 = A_Hq(n, sx_singlet, nf, L)
    A_hg_3 = A_Hg(n, sx_singlet, nf, L)

    A_gq_3 = A_gq(n, sx_singlet, nf, L)
    A_gg_3 = A_gg(n, sx_singlet, nf, L)

    A_qq_ps_3 = A_qqPS(n, sx_singlet, nf, L)
    A_qq_ns_3 = A_qqNS(n, sx_non_singlet, nf, L)
    A_qg_3 = A_qg(n, sx_singlet, nf, L)

    A_S = np.array(
        [
            [A_gg_3, A_gq_3, 0.0],
            [A_qg_3, A_qq_ps_3 + A_qq_ns_3, 0.0],
            [A_hg_3, A_hq_3, 0.0],
        ],
        np.complex_,
    )
    return A_S


@nb.njit(cache=True)
def A_ns(n, sx_all, nf, L):
    r"""Computes the |N3LO| non-singlet |OME|.

    .. math::
        A^{NS,(3)} = \left(\begin{array}{cc}
            A_{qq,H}^{NS,(3)} & 0\\
            0 & 0\\
        \end{array}\right)

    When using the code, please cite the complete list of references available
    at the top of this module :mod:`eko.matching_conditions.as3`.

    Parameters
    ----------
    n : complex
        Mellin moment
    sx_all : list
        harmonic sums cache containing:

        .. math ::
            [[S_1,S_{-1}],
            [S_2,S_{-2}],
            [S_{3}, S_{2,1}, S_{2,-1}, S_{-2,1}, S_{-2,-1}, S_{-3}],
            [S_{4}, S_{3,1}, S_{2,1,1}, S_{-2,-2}, S_{-3, 1}, S_{-4}],],
            [S_{5}, S_{-5}]

    nf : int
        number of active flavor below the threshold
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
    A_NS : numpy.ndarray
        |N3LO| non-singlet |OME| :math:`A^{NS,(3)}`

    See Also
    --------
    A_qqNS_3 : :math:`A_{qq,H}^{NS,(3))}`

    """
    return np.array([[A_qqNS(n, sx_all, nf, L), 0.0], [0 + 0j, 0 + 0j]], np.complex_)



# -*- coding: utf-8 -*-
r"""
This module contains the QCD gamma function coefficients.

See :doc:`pQCD ingredients </theory/pQCD>`.
"""
import numba as nb

from .harmonics.constants import zeta3, zeta4, zeta5


@nb.njit(cache=True)
def gamma_qcd_as1():
    """
    Computes the first coefficient of the QCD gamma function.

    Implements Eq. (15) of :cite:`Vermaseren:1997fq`.

    Returns
    -------
        gamma_0 : float
            first coefficient of the QCD gamma function :math:`\\gamma_{m,0}^{n_f}`
    """
    return 4.0


@nb.njit(cache=True)
def gamma_qcd_as2(nf):
    """
    Computes the second coefficient of the QCD gamma function.

    Implements Eq. (15) of :cite:`Vermaseren:1997fq`.

    Parameters
    ----------
        nf : int
            number of active flavors

    Returns
    -------
        gamma_1 : float
            second coefficient of the QCD gamma function :math:`\\gamma_{m,1}^{n_f}`
    """
    return 202.0 / 3.0 - 20.0 / 9.0 * nf


@nb.njit(cache=True)
def gamma_qcd_as3(nf):
    """
    Computes the third coefficient of the QCD gamma function.

    Implements Eq. (15) of :cite:`Vermaseren:1997fq`.

    Parameters
    ----------
        nf : int
            number of active flavors

    Returns
    -------
        gamma_2 : float
            third coefficient of the QCD gamma function :math:`\\gamma_{m,2}^{n_f}`
    """
    return 1249.0 - (2216.0 / 27.0 + 160.0 / 3.0 * zeta3) * nf - 140.0 / 81.0 * nf**2


@nb.njit(cache=True)
def gamma_qcd_as4(nf):
    """
    Computes the fourth coefficient of the QCD gamma function.

    Implements Eq. (15) of :cite:`Vermaseren:1997fq`.

    Parameters
    ----------
        nf : int
            number of active flavors

    Returns
    -------
        gamma_3 : float
            fourth coefficient of the QCD gamma function :math:`\\gamma_{m,3}^{n_f}`
    """
    return (
        4603055.0 / 162.0
        + 135680.0 * zeta3 / 28.0
        - 8800.0 * zeta5
        + (
            -91723.0 / 27.0
            - 34192.0 * zeta3 / 9.0
            + 880.0 * zeta4
            + 18400.0 * zeta5 / 9.0
        )
        * nf
        + (5242.0 / 243.0 + 800.0 * zeta3 / 9.0 - 160.0 * zeta4 / 3.0) * nf**2
        + (332.0 / 243.0 + 64.0 * zeta3 / 27.0) * nf**3
    )


@nb.njit(cache=True)
def gamma(order, nf):
    """
    Compute the value of a gamma coefficient

    Parameters
    ----------
        order: int
            perturbative order
        nf : int
            number of active flavors

    returns
    -------
        gamma: float
            QCD gamma function coefficient
    """
    _gamma = 0.0

    if order == 1:
        _gamma = gamma_qcd_as1()
    elif order == 2:
        _gamma = gamma_qcd_as2(nf)
    elif order == 3:
        _gamma = gamma_qcd_as3(nf)
    elif order == 4:
        _gamma = gamma_qcd_as4(nf)
    else:
        raise ValueError("QCD Gamma coefficients beyond N3LO are not implemented!")
    return _gamma



# -*- coding: utf-8 -*-
"""This file implements the |N3LO| evolution integrals"""

import numba as nb
import numpy as np


@nb.njit(cache=True)
def roots(b_list):
    """Returns the roots of:

    .. math ::
        1 + b_1 a_s + b_2 a_s^2 + b_3 a_s^3 = 0

    Parameters
    ----------
    b_list : list
        :math:`[b_1, b_2, b_3]`
    Returns
    -------
    list
        list of complex roots

    """
    b1, b2, b3 = b_list
    d1 = -(b2**2) + 3 * b1 * b3
    d2 = -2 * b2**3 + 9 * b1 * b2 * b3 - 27 * b3**2
    d3 = ((d2 + np.sqrt(4 * d1**3 + d2**2)) / 2) ** (1.0 / 3.0)
    r1 = 1 / (3 * b3) * (-b2 - d1 / d3 + d3)
    r2 = (1 / (3 * b3)) * (
        -b2
        + ((1 + 1.0j * np.sqrt(3)) * d1) / (2 * d3)
        - ((1 - 1.0j * np.sqrt(3)) * d3) / 2
    )
    r3 = (1 / (3 * b3)) * (
        -b2
        + ((1 - 1.0j * np.sqrt(3)) * d1) / (2 * d3)
        - ((1 + 1.0j * np.sqrt(3)) * d3) / 2
    )
    return [r1, r2, r3]


@nb.njit(cache=True)
def derivative(r, b_list):
    r"""Returns the derivative:

    .. math ::
        \frac{d}{d a_s}(1 + b_1 a_s + b_2 a_s^2 + b_3 a_s^3) = b_1 + 2 b_2 r + 3 b_3 r^2

    Parameters
    ----------
    a_s :
        coupling constant
    b_list : list
        :math:`[b_1, b_2, b_3]`

    Returns
    -------
    float :
        evaluated derivative

    """
    b1, b2, b3 = b_list
    return b1 + 2 * b2 * r + 3 * b3 * r**2


@nb.njit(cache=True)
def j33_exact(a1, a0, beta0, b_list, roots):
    r"""|N3LO|-|N3LO| exact evolution definite integral
    evaluated at :math:`a_s-a_s^0`.

    .. math::

        j^{(3,3)}(a_s) =  \frac{1}{\beta_0} \sum_{r=r_1}^{r_3} \frac{\ln(a_s-r) r^2}{b_1 + 2 b_2 r + 3 b_3 r^2}

    Parameters
    ----------
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    beta0 : float
        :math:`\beta_0`
    b_list : list
        :math:`[b_1, b_2, b_3]`
    roots: :math:`[r_1,r_2,r_3]`
        list of roots of :math:`1 + b_1 a_s + b_2 a_s^2 + b_3 a_s^3`

    Returns
    -------
    float
        evaluated integral

    """
    integral = 0
    for r in roots:
        integral += r**2 / derivative(r, b_list) * np.log((a1 - r) / (a0 - r))
    return (1 / beta0) * integral


@nb.njit(cache=True)
def j23_exact(a1, a0, beta0, b_list, roots):
    r"""|NNLO|-|N3LO| exact evolution definite integral
    evaluated at :math:`a_s-a_s^0`.

    .. math::

        j^{(2,3)}(a_s) =  \frac{1}{\beta_0} \sum_{r=r_1}^{r_3} \frac{\ln(a_s-r) r}{b_1 + 2 b_2 r + 3 b_3 r^2}

    Parameters
    ----------
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    beta0 : float
        :math:`\beta_0`
    b_list : list
        :math:`[b_1, b_2, b_3]`
    roots: :math:`[r_1,r_2,r_3]`
        list of roots of :math:`1 + b_1 a_s + b_2 a_s^2 + b_3 a_s^3`

    Returns
    -------
    float
        evaluated integral

    """
    integral = 0
    for r in roots:
        integral += r / derivative(r, b_list) * np.log((a1 - r) / (a0 - r))
    return (1 / beta0) * integral


@nb.njit(cache=True)
def j13_exact(a1, a0, beta0, b_list, roots):
    r"""|NLO|-|N3LO| exact evolution definite integral
    evaluated at :math:`a_s-a_s^0`.

    .. math::

        j^{(1,3)}(a_s) = \frac{1}{\beta_0} \sum_{r=r_1}^{r_3} \frac{\ln(a_s-r)}{b_1 + 2 b_2 r + 3 b_3 r^2}

    Parameters
    ----------
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    beta0 : float
        :math:`\beta_0`
    b_list : list
        :math:`[b_1, b_2, b_3]`
    roots: :math:`[r_1,r_2,r_3]`
        list of roots of :math:`1 + b_1 a_s + b_2 a_s^2 + b_3 a_s^3`

    Returns
    -------
    float
        evaluated integral

    """
    integral = 0
    for r in roots:
        integral += 1.0 / derivative(r, b_list) * np.log((a1 - r) / (a0 - r))
    return (1 / beta0) * integral


@nb.njit(cache=True)
def j03_exact(j00, j13, j23, j33, b_list):
    r"""|LO|-|N3LO| exact evolution definite integral
    evaluated at :math:`a_s-a_s^0`.

    .. math::

        j^{(0,3)}(a_s) = \frac{1}{\beta_0} \left( \ln(a) - \sum_{r=r_1}^{r_3} \frac{b_1 \ln(a_s-r) + b_2 \ln(a_s-r) r + b_3 \ln(a_s-r) r^2}{b_1 + 2 b_2 r + 3 b_3 r^2} \right )

    Parameters
    ----------
    j00: float
        |LO|-|LO| evolution integral
    j13: float
        |NLO|-|N3LO| evolution integral
    j23: float
        |NNLO|-|N3LO| evolution integral
    j33: float
        |N3LO|-|N3LO| evolution integral
    b_list : list
        :math:`[b_1, b_2, b_3]`

    Returns
    -------
    float
        evaluated integral

    """
    b1, b2, b3 = b_list
    return j00 - b1 * j13 - b2 * j23 - b3 * j33


@nb.njit(cache=True)
def j33_expanded(a1, a0, beta0):
    r"""|N3LO|-|N3LO| expanded evolution definite integral
    evaluated at :math:`a_s-a_s^0`.

    .. math::
        j^{(3,3)}_{exp}(a_s) = \frac{1}{3 \beta_0} a_s^3

    Parameters
    ----------
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    beta0 : float
        :math:`\beta_0`

    Returns
    -------
    float
        evaluated integral

    """
    return 1 / (3 * beta0) * (a1**3 - a0**3)


@nb.njit(cache=True)
def j23_expanded(a1, a0, beta0, b_list):
    r"""|NNLO|-|N3LO| expanded evolution definite integral
    evaluated at :math:`a_s-a_s^0`.

    .. math::
        j^{(2,3)}_{exp}(a_s) = \frac{1}{\beta_0} ( \frac{1}{2} a_s^2 - \frac{b_1}{3} as^3)

    Parameters
    ----------
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    beta0 : float
        :math:`\beta_0`
    b_list : list
        :math:`[b_1, b_2, b_3]`

    Returns
    -------
    float
        evaluated integral

    """
    b1 = b_list[0]
    return 1 / beta0 * (1 / 2 * (a1**2 - a0**2) - b1 / 3 * (a1**3 - a0**3))


@nb.njit(cache=True)
def j13_expanded(a1, a0, beta0, b_list):
    r"""|NLO|-|N3LO| expanded evolution definite integral
    evaluated at :math:`a_s-a_s^0`.

    .. math::
        j^{(1,3)}_{exp}(a_s) = \frac{1}{\beta_0} ( a_s - \frac{b_1}{2} a_s^2 + \frac{b_1^2-b_2}{3} as^3)

    Parameters
    ----------
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    beta0 : float
        :math:`\beta_0`
    b_list : list
        :math:`[b_1, b_2, b_3]`

    Returns
    -------
    float
        evaluated integral

    """
    b1, b2, _ = b_list
    return (1 / beta0) * (
        (a1 - a0)
        - b1 / 2 * (a1**2 - a0**2)
        + (b1**2 - b2) / 3 * (a1**3 - a0**3)
    )


@nb.njit(cache=True)
def j03_expanded(j00, j13, j23, j33, b_list):
    r"""|LO|-|N3LO| expanded evolution definite integral
    evaluated at :math:`a_s-a_s^0`.

    .. math::
        j^{(0,3)}_{exp}(a_s) = j^{(0,0)} - b_1 j^{(1,3)}_{exp}(a_s) - b_2 j^{(2,3)}_{exp}(a_s) - b_3 j^{(3,3)}_{exp}(a_s)

    Parameters
    ----------
    j00: float
        |LO|-|LO| evolution integral
    j13: float
        |NLO|-|N3LO| expanded evolution integral
    j23: float
        |NNLO|-|N3LO| expanded evolution integral
    j33: float
        |N3LO|-|N3LO| expanded evolution integral
    b_list : list
        :math:`[b_1, b_2, b_3]`

    Returns
    -------
    float
        evaluated integral

    See Also
    --------
    j03_exact

    """
    return j03_exact(j00, j13, j23, j33, b_list)



# -*- coding: utf-8 -*-
r"""
The write-up of the matching conditions is given in
:doc:`Matching Conditions </theory/Matching>`.

"""

import numpy as np

from .. import basis_rotation as br


def pids_from_intrinsic_evol(label, nf, normalize):
    r"""
    Obtain the list of pids with their corresponding weight, that are contributing to ``evol``

    The normalization of the weights is only needed for the output rotation:

    - if we want to build e.g. the singlet in the initial state we simply have to sum
      to obtain :math:`S = u + \bar u + d + \bar d + \ldots`
    - if we want to rotate back in the output we have to *normalize* the weights:
      e.g. in nf=3 :math:`u = \frac 1 6 S + \frac 1 6 V + \ldots`

    The normalization can only happen here since we're actively cutting out some
    flavor (according to ``nf``).

    Parameters
    ----------
        evol : str
            evolution label
        nf : int
            maximum number of light flavors
        normalize : bool
            normalize output

    Returns
    -------
        m : list
    """
    try:
        evol_idx = br.evol_basis.index(label)
        is_evol = True
    except ValueError:
        is_evol = False
    if is_evol:
        weights = br.rotate_flavor_to_evolution[evol_idx].copy()
        for j, pid in enumerate(br.flavor_basis_pids):
            if nf < abs(pid) <= 6:
                weights[j] = 0
    else:
        weights = rotate_pm_to_flavor(label)
    # normalize?
    if normalize:
        norm = weights @ weights
        weights = weights / norm
    return weights


def get_range(evol_labels):
    """
    Determine the number of light and heavy flavors participating in the input and output.

    Here, we assume that the T distributions (e.g. T15) appears *always*
    before the corresponding V distribution (e.g. V15).

    Returns
    -------
        nf_in : int
            number of light flavors in the input
        nf_out : int
            number of light flavors in the output
    """
    nf_in = 3
    nf_out = 3

    def update(label):
        nf = 3
        if label[0] == "T":
            nf = round(np.sqrt(int(label[1:]) + 1))
        return nf

    for op in evol_labels:
        nf_in = max(update(op.input), nf_in)
        nf_out = max(update(op.target), nf_out)

    return nf_in, nf_out


def rotate_pm_to_flavor(label):
    """
    Rotate from +- basis to flavor basis.

    Parameters
    ----------
        label : str
            label

    Returns
    -------
        l : list(float)
            list of weights
    """
    # g and ph are unaltered
    if label in ["g", "ph"]:
        return br.rotate_flavor_to_evolution[br.evol_basis.index(label)].copy()
    # no it has to be a quark with + or - appended
    if label[0] not in br.quark_names or label[1] not in ["+", "-"]:
        raise ValueError(f"Invalid pm label: {label}")
    l = np.zeros(len(br.flavor_basis_pids))
    idx = br.flavor_basis_names.index(label[0])
    pid = br.flavor_basis_pids[idx]
    l[idx] = 1
    # + is +, - is -
    if label[1] == "+":
        l[br.flavor_basis_pids.index(-pid)] = 1
    else:
        l[br.flavor_basis_pids.index(-pid)] = -1
    return l


def rotate_matching(nf, inverse=False):
    """
    Rotation between matching basis (with e.g. S,g,...V8 and c+,c-) and new true evolution basis
    (with S,g,...V8,T15,V15).

    Parameters
    ----------
        nf : int
            number of active flavors in the higher patch: to activate T15, nf=4
        inverse : bool
            use inverse conditions?

    Returns
    -------
        l : dict
            mapping in dot notation between the bases
    """
    # the gluon and the photon do not care about new quarks
    l = {"g.g": 1.0, "ph.ph": 1.0}
    # already active distributions
    for k in range(2, nf):  # nf is the upper, so excluded
        n = k**2 - 1
        l[f"V{n}.V{n}"] = 1.0
        l[f"T{n}.T{n}"] = 1.0
    # the new contributions
    n = nf**2 - 1  # nf is pointing upwards
    q = br.quark_names[nf - 1]
    for (tot, oth, qpm) in (("S", f"T{n}", f"{q}+"), ("V", f"V{n}", f"{q}-")):
        if inverse:
            l[f"{tot}.{tot}"] = (nf - 1.0) / nf
            l[f"{tot}.{oth}"] = 1.0 / nf
            l[f"{qpm}.{tot}"] = 1.0 / nf
            l[f"{qpm}.{oth}"] = -1.0 / nf
        else:
            l[f"{tot}.{tot}"] = 1.0
            l[f"{tot}.{qpm}"] = 1.0
            l[f"{oth}.{tot}"] = 1.0
            l[f"{oth}.{qpm}"] = -(nf - 1.0)
    # also higher quarks do not care
    for k in range(nf + 1, 6 + 1):
        q = br.quark_names[k - 1]
        for sgn in "+-":
            l[f"{q}{sgn}.{q}{sgn}"] = 1.0
    return l


def rotate_matching_inverse(nf):
    return rotate_matching(nf, True)


def pids_from_intrinsic_unified_evol(label, nf, normalize):
    r"""
    Obtain the list of pids with their corresponding weight, that are contributing to intrinsic
    unified evolution.

    Parameters
    ----------
        evol : str
            evolution label
        nf : int
            maximum number of light flavors
        normalize : bool
            normalize output

    Returns
    -------
        m : list
    """
    if label in ["ph", "g", "S", "V"]:
        return pids_from_intrinsic_evol(label, nf, normalize)
    if label[0] in br.quark_names[3:]:
        weights = rotate_pm_to_flavor(label)
    else:
        weights = np.array([0.0] * len(br.flavor_basis_pids))
        mapping = {
            "d3": {1: 1.0, 3: -1.0},  # T3d = d+ - s+
            "d8": {1: 1.0, 3: 1.0, 5: -2.0},  # T8d = d+ + s+ - 2b+
            "u3": {2: 1.0, 4: -1.0},  # T3u = u+ - c+
            "u8": {2: 1.0, 4: 1.0, 6: -2.0},  # T8u = u+ + c+ - 2t+
            "delta": {
                3: {2: 2.0, 1: -1.0, 3: -1.0},  # Sdelta = 2u+ - d+ -s+
                4: {2: 1.0, 4: 1.0, 1: -1.0, 3: -1.0},  # Sdelta = u+ + c+ - d+ - s+
                5: {
                    2: 3.0 / 2.0,
                    4: 3.0 / 2.0,
                    1: -1.0,
                    3: -1.0,
                    5: -1.0,
                },  # Sdelta = 3/2u+ + 3/2c+ - d+ - s+ - b+
                6: {
                    2: 1.0,
                    4: 1.0,
                    6: 1.0,
                    1: -1.0,
                    3: -1.0,
                    5: -1.0,
                },  # Sdelta = u+ + c+ + t+ - d+ -s+ - b+
            },
        }
        cur_map = mapping[label[1:]]
        if label[1:] == "delta":
            cur_map = cur_map[nf]
        for q, w in cur_map.items():
            weights[br.flavor_basis_pids.index(q)] = w
            weights[br.flavor_basis_pids.index(-q)] = (
                -1 if label[0] == "V" else 1.0
            ) * w

    # normalize?
    if normalize:
        norm = weights @ weights
        weights = weights / norm
    return weights



# -*- coding: utf-8 -*-
"""
This module defines the CLI - see :doc:`here </code/genpdf>`
"""

import click

from . import generate_pdf, install_pdf


@click.group()
def cli():
    pass


@cli.command("generate")
@click.argument("name")
@click.argument("labels", nargs=-1)
@click.option("-p", "--parent-pdf-set", default=None, help="parent pdf set")
@click.option("-m", "--members", is_flag=True, help="generate all the members")
@click.option("-i", "--install", is_flag=True, help="install into LHAPDF")
def cli_generate_pdf(name, labels, parent_pdf_set, members, install):
    """Generate a new PDF from a parent set with given flavors"""
    if len(labels) == 0:
        raise ValueError("Labels must contain at least one element")
    return generate_pdf(name, labels, parent_pdf_set, members, None, install)


@cli.command("install")
@click.argument("name")
def cli_install_pdf(name):
    """Install the PDF on LHAPDF directory"""
    return install_pdf(name)



# -*- coding: utf-8 -*-
import copy


def update(theory, operators):
    """
    Upgrade the legacy theory and observable runcards with the new settings.

    Parameters
    ----------
        theory : dict
            theory runcard
        observables : dict
            observable runcard

    Returns
    -------
        new_theory : dict
            upgraded theory runcard
        new_obs : dict
            upgraded observable runcard
    """
    new_theory = copy.deepcopy(theory)
    new_operators = copy.deepcopy(operators)
    if "alphaqed" in new_theory:
        new_theory["alphaem"] = new_theory.pop("alphaqed")
    if "QED" in new_theory:
        new_theory["order"] = (new_theory.pop("PTO") + 1, new_theory.pop("QED"))
    if operators is not None:
        if isinstance(new_operators["ev_op_max_order"], int):
            new_operators["ev_op_max_order"] = (
                new_operators["ev_op_max_order"],
                new_theory["order"][1],
            )
    return new_theory, new_operators


def update_theory(theory):
    """
    Upgrade the legacy theory runcards with the new settings.

    Parameters
    ----------
        theory : dict
            theory runcard

    Returns
    -------
        new_theory : dict
            upgraded theory runcard
    """
    return update(theory, None)[0]



# -*- coding: utf-8 -*-
import numpy as np

from .. import basis_rotation as br
from .. import member
from . import flavors


class PhysicalOperator(member.OperatorBase):
    """
    This joins several fixed flavor scheme operators together.

    - provides the connection between the 7-dimensional anomalous dimension
      basis and the 15-dimensional evolution basis
    - provides the connection between the 15-dimensional evolution basis
      and the 169-dimensional flavor basis


    Parameters
    ----------
        op_members : dict
            list of all members
        q2_final : float
            final scale
    """

    @classmethod
    def ad_to_evol_map(cls, op_members, nf, q2_final, intrinsic_range):
        """
        Obtain map between the 3-dimensional anomalous dimension basis and the
        4-dimensional evolution basis.

        .. todo:: in VFNS sometimes IC is irrelevant if nf>=4

        Parameters
        ----------
            op_members : dict
                operator members in anomalous dimension basis
            nf : int
                number of active light flavors
            intrinsic_range : sequence
                intrinsic heavy flavors

        Returns
        -------
            m : dict
                map
        """
        # constant elements
        m = {
            "S.S": op_members[(100, 100)],
            "S.g": op_members[(100, 21)],
            "g.g": op_members[(21, 21)],
            "g.S": op_members[(21, 100)],
            "V.V": op_members[(br.non_singlet_pids_map["nsV"], 0)],
        }
        # add elements which are already active
        for f in range(2, nf + 1):
            n = f**2 - 1
            m[f"V{n}.V{n}"] = op_members[(br.non_singlet_pids_map["ns-"], 0)]
            m[f"T{n}.T{n}"] = op_members[(br.non_singlet_pids_map["ns+"], 0)]
        # deal with intrinsic heavy quark PDFs
        if intrinsic_range is not None:
            hqfl = "cbt"
            op_id = member.OpMember.id_like(
                op_members[(br.non_singlet_pids_map["nsV"], 0)]
            )
            for intr_fl in intrinsic_range:
                if intr_fl <= nf:  # light quarks are not intrinsic
                    continue
                hq = hqfl[intr_fl - 4]  # find name
                # intrinsic means no evolution, i.e. they are evolving with the identity
                m[f"{hq}+.{hq}+"] = op_id.copy()
                m[f"{hq}-.{hq}-"] = op_id.copy()
        # map key to MemberName
        return cls.promote_names(m, q2_final)

    def to_flavor_basis_tensor(self):
        """
        Convert the computations into an rank 4 tensor over flavor operator space and
        momentum fraction operator space.

        Returns
        -------
            tensor : numpy.ndarray
                EKO
        """
        nf_in, nf_out = flavors.get_range(self.op_members.keys())
        len_pids = len(br.flavor_basis_pids)
        len_xgrid = list(self.op_members.values())[0].value.shape[0]
        # dimension will be pids^2 * xgrid^2
        value_tensor = np.zeros((len_pids, len_xgrid, len_pids, len_xgrid))
        error_tensor = value_tensor.copy()
        for name, op in self.op_members.items():
            in_pids = flavors.pids_from_intrinsic_evol(name.input, nf_in, False)
            out_pids = flavors.pids_from_intrinsic_evol(name.target, nf_out, True)
            for out_idx, out_weight in enumerate(out_pids):
                for in_idx, in_weight in enumerate(in_pids):
                    # keep the outer index to the left as we're multiplying from the right
                    value_tensor[
                        out_idx,  # output pid (position)
                        :,  # output momentum fraction
                        in_idx,  # input pid (position)
                        :,  # input momentum fraction
                    ] += out_weight * (op.value * in_weight)
                    error_tensor[
                        out_idx,  # output pid (position)
                        :,  # output momentum fraction
                        in_idx,  # input pid (position)
                        :,  # input momentum fraction
                    ] += out_weight * (op.error * in_weight)
        return value_tensor, error_tensor



# -*- coding: utf-8 -*-
import copy

import numpy as np

from eko import basis_rotation as br


def pid_to_flavor(pids):
    """
    Create flavor representations from PIDs.

    Parameters
    ----------
        pids : list(int)
            active PIDs

    Returns
    -------
        numpy.ndarray
            list of reprentations for each PID
    """
    ps = []
    zeros = np.zeros(len(br.flavor_basis_pids))
    for pid in pids:
        p = zeros.copy()
        idx = br.flavor_basis_pids.index(pid)
        p[idx] = 1.0
        ps.append(p)
    return np.array(ps)


def evol_to_flavor(labels):
    """
    Create flavor representations from evolution members.

    Parameters
    ----------
        labels : list(str)
            active evolution distributions

    Returns
    -------
        numpy.ndarray
            list of reprentations for each distribution
    """
    ps = []
    for label in labels:
        idx = br.evol_basis.index(label)
        ps.append(br.rotate_flavor_to_evolution[idx].copy())
    return np.array(ps)


def project(blocks, reprs):
    """
    Project some combination of flavors defined by reprs from the blocks.

    Parameters
    ----------
        blocks : list(dict)
            PDF blocks
        reprs : list(int)
            active distributions in flavor representation

    Returns
    -------
        list(dict) :
            filtered blocks
    """
    new_blocks = copy.deepcopy(blocks)
    for block in new_blocks:
        current_pids = block["pids"]
        current_data = block["data"].T
        if len(current_data) == 0:
            continue
        # load all flavors
        flavor_data = [np.zeros_like(current_data[0]) for pid in br.flavor_basis_pids]
        for pid, pdf in zip(current_pids, current_data):
            idx = br.flavor_basis_pids.index(pid)
            flavor_data[idx] = pdf
        flavor_data = np.array(flavor_data)
        new_data = np.zeros_like(flavor_data)
        for elem in reprs:
            proj = elem[:, np.newaxis] * elem
            new_data += proj @ flavor_data / (elem @ elem)
        block["pids"] = br.flavor_basis_pids
        block["data"] = np.array(new_data).T
    return new_blocks


def is_evolution_labels(labels):
    """
    Check whether the labels are provided in evolution basis

    Parameters
    ----------
        labels : list()
            list of labels

    Returns
    -------
        bool :
            is evolution basis
    """
    for label in labels:
        if not isinstance(label, str):
            return False
        if label not in br.evol_basis:
            return False
    return True


def is_pid_labels(labels):
    """
    Check whether the labels are provided in flavor basis

    Parameters
    ----------
        labels : list()
            list of labels

    Returns
    -------
        bool :
            is flavor basis
    """
    try:
        labels = np.array(labels, dtype=np.int_)
    except (ValueError, TypeError):
        return False
    for label in labels:
        if not isinstance(label, np.int_):
            return False
        if label not in br.flavor_basis_pids:
            return False
    return True



# -*- coding: utf-8 -*-
r"""
This module contains the Altarelli-Parisi splitting kernels.

Normalization is given by

.. math::
    \mathbf{P}(x) = \sum\limits_{j=0} a_s^{j+1} \mathbf P^{(j)}(x)

with :math:`a_s = \frac{\alpha_S(\mu^2)}{4\pi}`.
The 3-loop references for the non-singlet :cite:`Moch:2004pa`
and singlet :cite:`Vogt:2004mw` case contain also the lower
order results. The results are also determined in Mellin space in
terms of the anomalous dimensions (note the additional sign!)

.. math::
    \gamma(N) = - \mathcal{M}[\mathbf{P}(x)](N)
"""

import numba as nb
import numpy as np

from .. import basis_rotation as br
from .. import harmonics
from . import aem1, aem2, as1, as1aem1, as2, as3, as4


@nb.njit(cache=True)
def exp_singlet(gamma_S):
    r"""
    Computes the exponential and the eigensystem of the singlet anomalous dimension matrix

    Parameters
    ----------
        gamma_S : numpy.ndarray
            singlet anomalous dimension matrix

    Returns
    -------
        exp : numpy.ndarray
            exponential of the singlet anomalous dimension matrix :math:`\gamma_{S}(N)`
        lambda_p : complex
            positive eigenvalue of the singlet anomalous dimension matrix
            :math:`\gamma_{S}(N)`
        lambda_m : complex
            negative eigenvalue of the singlet anomalous dimension matrix
            :math:`\gamma_{S}(N)`
        e_p : numpy.ndarray
            projector for the positive eigenvalue of the singlet anomalous
            dimension matrix :math:`\gamma_{S}(N)`
        e_m : numpy.ndarray
            projector for the negative eigenvalue of the singlet anomalous
            dimension matrix :math:`\gamma_{S}(N)`

    See Also
    --------
        eko.anomalous_dimensions.as1.gamma_singlet : :math:`\gamma_{S}^{(0)}(N)`
        eko.anomalous_dimensions.as2.gamma_singlet : :math:`\gamma_{S}^{(1)}(N)`
        eko.anomalous_dimensions.as3.gamma_singlet : :math:`\gamma_{S}^{(2)}(N)`
    """
    # compute eigenvalues
    det = np.sqrt(
        np.power(gamma_S[0, 0] - gamma_S[1, 1], 2) + 4.0 * gamma_S[0, 1] * gamma_S[1, 0]
    )
    lambda_p = 1.0 / 2.0 * (gamma_S[0, 0] + gamma_S[1, 1] + det)
    lambda_m = 1.0 / 2.0 * (gamma_S[0, 0] + gamma_S[1, 1] - det)
    # compute projectors
    identity = np.identity(2, np.complex_)
    c = 1.0 / det
    e_p = +c * (gamma_S - lambda_m * identity)
    e_m = -c * (gamma_S - lambda_p * identity)
    exp = e_m * np.exp(lambda_m) + e_p * np.exp(lambda_p)
    return exp, lambda_p, lambda_m, e_p, e_m


@nb.njit(cache=True)
def gamma_ns(order, mode, n, nf):
    r"""Computes the tower of the non-singlet anomalous dimensions

    Parameters
    ----------
    order : tuple(int,int)
        perturbative orders
    mode : 10201 | 10101 | 10200
        sector identifier
    n : complex
        Mellin variable
    nf : int
        Number of active flavors

    Returns
    -------
    numpy.ndarray
        non-singlet anomalous dimensions

    See Also
    --------
    eko.anomalous_dimensions.as1.gamma_ns : :math:`\gamma_{ns}^{(0)}(N)`
    eko.anomalous_dimensions.as2.gamma_nsp : :math:`\gamma_{ns,+}^{(1)}(N)`
    eko.anomalous_dimensions.as2.gamma_nsm : :math:`\gamma_{ns,-}^{(1)}(N)`
    eko.anomalous_dimensions.as3.gamma_nsp : :math:`\gamma_{ns,+}^{(2)}(N)`
    eko.anomalous_dimensions.as3.gamma_nsm : :math:`\gamma_{ns,-}^{(2)}(N)`
    eko.anomalous_dimensions.as3.gamma_nsv : :math:`\gamma_{ns,v}^{(2)}(N)`
    eko.anomalous_dimensions.as4.gamma_nsp : :math:`\gamma_{ns,+}^{(3)}(N)`
    eko.anomalous_dimensions.as4.gamma_nsm : :math:`\gamma_{ns,-}^{(3)}(N)`
    eko.anomalous_dimensions.as4.gamma_nsv : :math:`\gamma_{ns,v}^{(3)}(N)`

    """
    # cache the s-es
    if order[0] >= 4:
        full_sx_cache = harmonics.compute_cache(n, 5, is_singlet=False)
        sx = np.array(
            [
                full_sx_cache[0][0],
                full_sx_cache[1][0],
                full_sx_cache[2][0],
                full_sx_cache[3][0],
            ]
        )
    else:
        sx = harmonics.sx(n, max_weight=order[0] + 1)
    # now combine
    gamma_ns = np.zeros(order[0], np.complex_)
    gamma_ns[0] = as1.gamma_ns(n, sx[0])
    # NLO and beyond
    if order[0] >= 2:
        if mode == 10101:
            gamma_ns_1 = as2.gamma_nsp(n, nf, sx)
        # To fill the full valence vector in NNLO we need to add gamma_ns^1 explicitly here
        elif mode in [10201, 10200]:
            gamma_ns_1 = as2.gamma_nsm(n, nf, sx)
        else:
            raise NotImplementedError("Non-singlet sector is not implemented")
        gamma_ns[1] = gamma_ns_1
    # NNLO and beyond
    if order[0] >= 3:
        if mode == 10101:
            gamma_ns_2 = as3.gamma_nsp(n, nf, sx)
        elif mode == 10201:
            gamma_ns_2 = as3.gamma_nsm(n, nf, sx)
        elif mode == 10200:
            gamma_ns_2 = as3.gamma_nsv(n, nf, sx)
        gamma_ns[2] = gamma_ns_2
    # N3LO
    if order[0] >= 4:
        if mode == 10101:
            gamma_ns_3 = as4.gamma_nsp(n, nf, full_sx_cache)
        elif mode == 10201:
            gamma_ns_3 = as4.gamma_nsm(n, nf, full_sx_cache)
        elif mode == 10200:
            gamma_ns_3 = as4.gamma_nsv(n, nf, full_sx_cache)
        gamma_ns[3] = gamma_ns_3
    return gamma_ns


@nb.njit(cache=True)
def gamma_singlet(order, n, nf):
    r"""Computes the tower of the singlet anomalous dimensions matrices

    Parameters
    ----------
    order : tuple(int,int)
        perturbative orders
    n : complex
        Mellin variable
    nf : int
        Number of active flavors

    Returns
    -------
    numpy.ndarray
        singlet anomalous dimensions matrices

    See Also
    --------
    eko.anomalous_dimensions.as1.gamma_singlet : :math:`\gamma_{S}^{(0)}(N)`
    eko.anomalous_dimensions.as2.gamma_singlet : :math:`\gamma_{S}^{(1)}(N)`
    eko.anomalous_dimensions.as3.gamma_singlet : :math:`\gamma_{S}^{(2)}(N)`
    eko.anomalous_dimensions.as4.gamma_singlet : :math:`\gamma_{S}^{(3)}(N)`

    """
    # cache the s-es
    if order[0] >= 4:
        full_sx_cache = harmonics.compute_cache(n, 5, is_singlet=False)
        sx = np.array(
            [
                full_sx_cache[0][0],
                full_sx_cache[1][0],
                full_sx_cache[2][0],
                full_sx_cache[3][0],
            ]
        )
    elif order[0] >= 3:
        # here we need only S1,S2,S3,S4
        sx = harmonics.sx(n, max_weight=order[0] + 1)
    else:
        sx = harmonics.sx(n, max_weight=order[0])

    gamma_s = np.zeros((order[0], 2, 2), np.complex_)
    gamma_s[0] = as1.gamma_singlet(n, sx[0], nf)
    if order[0] >= 2:
        gamma_s[1] = as2.gamma_singlet(n, nf, sx)
    if order[0] >= 3:
        gamma_s[2] = as3.gamma_singlet(n, nf, sx)
    if order[0] >= 4:
        gamma_s[3] = as4.gamma_singlet(n, nf, full_sx_cache)
    return gamma_s



# -*- coding: utf-8 -*-
r"""Contains the QCD beta function coefficients and the handling of the running coupling :math:`\alpha_s`.

See :doc:`pQCD ingredients </theory/pQCD>`.
"""

import logging

import numba as nb
import numpy as np
import scipy

from . import constants, thresholds
from .beta import b_qcd, b_qed, beta_qcd, beta_qed

logger = logging.getLogger(__name__)


def couplings_mod_ev(mod_ev):
    """Map ModEv key to the available strong coupling evolution methods.

    Parameters
    ----------
    mod_ev : str
        evolution mode

    Returns
    -------
    str
        coupling mode
    """
    if mod_ev in ["EXA", "iterate-exact", "decompose-exact", "perturbative-exact"]:
        return "exact"
    if mod_ev in [
        "TRN",
        "truncated",
        "ordered-truncated",
        "EXP",
        "iterate-expanded",
        "decompose-expanded",
        "perturbative-expanded",
    ]:
        return "expanded"
    raise ValueError(f"Unknown evolution mode {mod_ev}")


@nb.njit(cache=True)
def exact_lo(ref, beta0, lmu):
    r"""Compute expanded solution at |LO|.

    Parameters
    ----------
    ref : float
        reference value of the coupling
    beta0 : float
        first coefficient of the beta function
    lmu : float
        logarithm of the ratio between target and reference scales

    Returns
    -------
    float
        coupling at target scale :math:`a(\mu_R^2)`
    """
    den = 1.0 + beta0 * ref * lmu
    return ref / den


@nb.njit(cache=True)
def expanded_nlo(ref, beta0, b1, lmu):
    r"""Compute expanded solution at |NLO|.

    Implement the default expression for |NLO| expanded solution, e.g. the one implemented in |APFEL|

    Parameters
    ----------
    ref : float
        reference value of the coupling
    beta0 : float
        first coefficient of the beta function
    b1 : float
        second coefficient of the b function
    lmu : float
        logarithm of the ratio between target and reference scales

    Returns
    -------
    float
        coupling at target scale :math:`a(\mu_R^2)`
    """
    den = 1.0 + beta0 * ref * lmu
    a_LO = exact_lo(ref, beta0, lmu)
    as_NLO = a_LO * (1 - b1 * a_LO * np.log(den))
    return as_NLO


@nb.njit(cache=True)
def expanded_nnlo(ref, beta0, b1, b2, lmu):
    r"""Compute expanded solution at |NNLO|.

    Implement the |NNLO| expanded solution from |APFEL| (not the default expression)

    Parameters
    ----------
    ref : float
        reference value of the coupling
    beta0 : float
        first coefficient of the beta function
    b1 : float
        second coefficient of the b function
    b2 : float
        third coefficient of the b function
    lmu : float
        logarithm of the ratio between target and reference scales

    Returns
    -------
    float
        coupling at target scale :math:`a(\mu_R^2)`
    """
    a_LO = exact_lo(ref, beta0, lmu)
    a_NLO = expanded_nlo(ref, beta0, b1, lmu)
    res = a_LO * (
        1.0 + a_LO * (a_LO - ref) * (b2 - b1**2) + a_NLO * b1 * np.log(a_NLO / ref)
    )
    return res


@nb.njit(cache=True)
def expanded_n3lo(ref, beta0, b1, b2, b3, lmu):
    r"""Compute expanded solution at |N3LO|.

    Implement the |N3LO| expanded solution obtained via iterated solution of the RGE :cite:`Rottoli`

    Parameters
    ----------
    ref : float
        reference value of the coupling
    beta0 : float
        first coefficient of the beta function
    b1 : float
        second coefficient of the b function
    b2 : float
        third coefficient of the b function
    b3 : float
        fourth coefficient of the b function
    lmu : float
        logarithm of the ratio between target and reference scales

    Returns
    -------
    float
        coupling at target scale :math:`a(\mu_R^2)`
    """
    a_LO = exact_lo(ref, beta0, lmu)
    log_fact = np.log(a_LO)
    res = expanded_nnlo(ref, beta0, b1, b2, lmu)
    res += (
        a_LO**4
        / (2 * beta0**3)
        * (
            -2 * b1**3 * np.log(ref) ** 3
            + 5 * b1**3 * log_fact**2
            + 2 * b1**3 * log_fact**3
            + b1**3 * np.log(ref) ** 2 * (5 + 6 * log_fact)
            + 2 * beta0 * b1 * log_fact * (b2 + 2 * (b1**2 - beta0 * b2) * lmu * ref)
            - beta0**2
            * lmu
            * ref
            * (
                -2 * b1 * b2
                + 2 * beta0 * b3
                + (b1**3 - 2 * beta0 * b1 * b2 + beta0**2 * b3) * lmu * ref
            )
            - 2
            * b1
            * np.log(ref)
            * (
                5 * b1**2 * log_fact
                + 3 * b1**2 * log_fact**2
                + beta0 * (b2 + 2 * (b1**2 - beta0 * b2) * lmu * ref)
            )
        )
    )
    return res


@nb.njit(cache=True)
def expanded_qcd(ref, order, nf, lmu):
    r"""Compute QCD expanded solution at a given order.

    Parameters
    ----------
    ref : float
        reference value of the strong coupling
    order : int
        QCD order
    nf : int
        number of flavors
    lmu : float
        logarithm of the ratio between target and reference scales

    Returns
    -------
    float
        strong coupling at target scale :math:`a_s(\mu_R^2)`
    """
    res_as = ref
    if order >= 1:
        beta_qcd0 = beta_qcd((2, 0), nf)
        # QCD LO
        as_LO = exact_lo(ref, beta_qcd0, lmu)
        res_as = as_LO
    # NLO
    if order >= 2:
        b_qcd1 = b_qcd((3, 0), nf)
        as_NLO = expanded_nlo(ref, beta_qcd0, b_qcd1, lmu)
        res_as = as_NLO
    # NNLO
    if order >= 3:
        b_qcd2 = b_qcd((4, 0), nf)
        as_NNLO = expanded_nnlo(ref, beta_qcd0, b_qcd1, b_qcd2, lmu)
        res_as = as_NNLO
    # N3LO
    if order >= 4:
        b_qcd3 = b_qcd((5, 0), nf)
        as_N3LO = expanded_n3lo(
            ref,
            beta_qcd0,
            b_qcd1,
            b_qcd2,
            b_qcd3,
            lmu,
        )
        res_as = as_N3LO
    return res_as


@nb.njit(cache=True)
def expanded_qed(ref, order, nf, lmu):
    r"""Compute QED expanded solution at a given order.

    Parameters
    ----------
    ref : float
        reference value of the QED coupling
    order : int
        QED order
    nf : int
        number of flavors
    lmu : float
            logarithm of the ratio between target and reference scales

    Returns
    -------
    float
        QED coupling at target scale :math:`a_em(\mu_R^2)`
    """
    res_aem = ref
    if order >= 1:
        beta_qed0 = beta_qed((0, 2), nf)
        # QED LO
        aem_LO = exact_lo(ref, beta_qed0, lmu)
        res_aem = aem_LO
    # NLO
    if order >= 2:
        b_qed1 = b_qed((0, 3), nf)
        aem_NLO = expanded_nlo(ref, beta_qed0, b_qed1, lmu)
        res_aem = aem_NLO
    return res_aem


@nb.njit(cache=True)
def couplings_expanded(order, couplings_ref, nf, scale_from, scale_to):
    r"""Compute expanded expression.

    Parameters
    ----------
    order : tuple(int, int)
        perturbation order
    couplings_ref : numpy.ndarray
        reference alpha_s and alpha
    nf : int
        value of nf for computing the couplings
    scale_from : float
        reference scale
    scale_to : float
        target scale

    Returns
    -------
    numpy.ndarray
        couplings at target scale :math:`a(\mu_R^2)`
    """
    # common vars
    lmu = np.log(scale_to / scale_from)
    res_as = expanded_qcd(couplings_ref[0], order[0], nf, lmu)
    res_aem = expanded_qed(couplings_ref[1], order[1], nf, lmu)
    if order[0] >= 2 and order[1] >= 2:
        beta_qcd0 = beta_qcd((2, 0), nf)
        beta_qed0 = beta_qed((0, 2), nf)
        res_as += (
            -couplings_ref[0] ** 2
            * b_qcd((2, 1), nf)
            * np.log(1 + beta_qcd0 * couplings_ref[1] * lmu)
        )
        res_aem += (
            -couplings_ref[1] ** 2
            * b_qed((1, 2), nf)
            * np.log(1 + beta_qed0 * couplings_ref[0] * lmu)
        )
    return np.array([res_as, res_aem])


class Couplings:
    r"""Compute the strong and electromagnetic coupling constants :math:`a_s, a_{em}`.

    Note that

    - all scale parameters (``scale_ref`` and ``scale_to``),
      have to be given as squared values, i.e. in units of :math:`\text{GeV}^2`
    - although, we only provide methods for
      :math:`a_i = \frac{\alpha_i(\mu^2)}{4\pi}` the reference value has to be
      given in terms of :math:`\alpha_i(\mu_0^2)` due to legacy reasons
    - the ``order`` refers to the perturbative order of the beta function, thus
      ``order=(0,0)`` means leading order beta function, means evolution with :math:`\beta_qcd_as2`,
      means running at 1-loop - so there is a natural mismatch between ``order`` and the
      number of loops by one unit

    Normalization is given by :cite:`Herzog:2017ohr`:

    .. math::
        \frac{da_s(\mu^2)}{d\ln\mu^2} = \beta(a_s) \
        = - \sum\limits_{n=0} \beta_n a_s^{n+2}(\mu^2) \quad
        \text{with}~ a_s = \frac{\alpha_s(\mu^2)}{4\pi}

    See :doc:`pQCD ingredients </theory/pQCD>`.

    Parameters
    ----------
    couplings_ref : numpy.ndarray
        alpha_s and \alpha(!) at the reference scale :math:`\alpha_s(\mu_0^2),\alpha(\mu_0^2)`
    scale_ref : float
        reference scale :math:`\mu_0^2`
    masses : list(float)
        list with quark masses squared
    thresholds_ratios : list(float)
        list with ratios between the mass and the matching scales squared
    order: tuple(int,int)
        Evaluated order of the beta function: ``0`` = LO, ...
    method : ["expanded", "exact"]
        Applied method to solve the beta function
    nf_ref : int
        if given, the number of flavors at the reference scale
    max_nf : int
        if given, the maximum number of flavors
    """

    def __init__(
        self,
        couplings_ref,
        scale_ref,
        masses,
        thresholds_ratios,
        order=(1, 0),
        method="exact",
        nf_ref=None,
        max_nf=None,
        hqm_scheme="POLE",
    ):
        # Sanity checks
        if couplings_ref[0] <= 0:
            raise ValueError(f"alpha_s_ref has to be positive - got {couplings_ref[0]}")
        if couplings_ref[1] <= 0:
            raise ValueError(
                f"alpha_em_ref has to be positive - got {couplings_ref[1]}"
            )
        if scale_ref <= 0:
            raise ValueError(f"scale_ref has to be positive - got {scale_ref}")
        if order[0] not in [0, 1, 2, 3, 4]:
            raise NotImplementedError("a_s beyond N3LO is not implemented")
        if order[1] not in [0, 1, 2]:
            raise NotImplementedError("a_em beyond NLO is not implemented")
        self.order = order
        if method not in ["expanded", "exact"]:
            raise ValueError(f"Unknown method {method}")
        self.method = method

        # create new threshold object
        self.a_ref = couplings_ref.copy() / 4.0 / np.pi  # convert to a_s and a_em
        self.thresholds = thresholds.ThresholdsAtlas(
            masses,
            scale_ref,
            nf_ref,
            thresholds_ratios=thresholds_ratios,
            max_nf=max_nf,
        )
        self.hqm_scheme = hqm_scheme
        logger.info(
            "Strong Coupling: a_s(µ_R^2=%f)%s=%f=%f/(4π)\nElectromagnetic Coupling: a_em(µ_R^2=%f)%s=%f=%f/(4π)",
            self.q2_ref,
            f"^(nf={nf_ref})" if nf_ref else "",
            self.a_ref[0],
            self.a_ref[0] * 4 * np.pi,
            self.q2_ref,
            f"^(nf={nf_ref})" if nf_ref else "",
            self.a_ref[1],
            self.a_ref[1] * 4 * np.pi,
        )
        # cache
        self.cache = {}

    @property
    def q2_ref(self):
        """Return reference scale."""
        return self.thresholds.q2_ref

    @classmethod
    def from_dict(cls, theory_card, masses=None):
        r"""Create object from theory dictionary.

        Parameters
        ----------
        theory_card : dict
            theory dictionary
        masses : list
            list of |MSbar| masses squared or None if POLE masses are used

        Returns
        -------
        Couplings
            created object
        """
        # read my values
        # TODO cast to a_s here
        alphas_ref = theory_card["alphas"]
        alphaem_ref = theory_card["alphaem"]
        couplings_ref = np.array([alphas_ref, alphaem_ref])
        nf_ref = theory_card["nfref"]
        q2_alpha = pow(theory_card["Qref"], 2)
        order = theory_card["order"]
        method = couplings_mod_ev(theory_card["ModEv"])
        hqm_scheme = theory_card["HQ"]
        if hqm_scheme not in ["MSBAR", "POLE"]:
            raise ValueError(f"{hqm_scheme} is not implemented, choose POLE or MSBAR")
        # adjust factorization scale / renormalization scale
        fact_to_ren = theory_card["fact_to_ren_scale_ratio"]
        heavy_flavors = "cbt"
        if masses is None:
            masses = np.power(
                [theory_card[f"m{q}"] / fact_to_ren for q in heavy_flavors], 2
            )
        else:
            masses = masses / fact_to_ren**2
        thresholds_ratios = np.power(
            [theory_card[f"k{q}Thr"] for q in heavy_flavors], 2
        )
        max_nf = theory_card["MaxNfAs"]
        return cls(
            couplings_ref,
            q2_alpha,
            masses,
            thresholds_ratios,
            order,
            method,
            nf_ref,
            max_nf,
            hqm_scheme,
        )

    def compute_exact(self, a_ref, nf, scale_from, scale_to):
        """Compute couplings via |RGE|.

        Parameters
        ----------
        as_ref : numpy.ndarray
            reference alpha_s and alpha
        nf : int
            value of nf for computing alpha_i
        scale_from : float
            reference scale
        scale_to : float
            target scale

        Returns
        -------
        numpy.ndarray
            couplings at target scale :math:`a(Q^2)`
        """
        # in LO fallback to expanded, as this is the full solution
        u = np.log(scale_to / scale_from)

        def unidimensional_exact(beta0, b_vec, u, a_ref, method, rtol):
            def rge(_t, a, b_vec):
                rge = -(a**2) * (np.sum([a**k * b for k, b in enumerate(b_vec)]))
                return rge

            res = scipy.integrate.solve_ivp(
                rge,
                (0, beta0 * u),
                (a_ref,),
                args=[b_vec],
                method=method,
                rtol=rtol,
            )
            return res.y[0][-1]

        if self.order in [(0, 0), (0, 1), (1, 0), (1, 1)]:
            return couplings_expanded(
                self.order, a_ref, nf, scale_from, float(scale_to)
            )
        if self.order[0] in [0, 1]:
            # return expanded solution for a_s and exact for a_em
            a_s = couplings_expanded(
                self.order, a_ref, nf, scale_from, float(scale_to)
            )[0]
            beta0_qed = beta_qed((0, 2), nf)
            b_qed_vec = [1.0]
            # NLO
            # if self.order[1] >= 2:  # I think that at this point this if is always true
            b_qed_vec.append(b_qed((0, 3), nf))
            a_em = unidimensional_exact(
                beta0_qed, b_qed_vec, u, a_ref[1], "Radau", 1e-6
            )
            return np.array([a_s, a_em])
        if self.order[1] in [0, 1]:
            # return expanded solution for a_em and exact for a_s
            a_em = couplings_expanded(
                self.order, a_ref, nf, scale_from, float(scale_to)
            )[1]
            beta0_qcd = beta_qcd((2, 0), nf)
            b_qcd_vec = [1.0]
            # NLO
            if self.order[0] >= 2:
                b_qcd_vec.append(b_qcd((3, 0), nf))
                # NNLO
                if self.order[0] >= 3:
                    b_qcd_vec.append(b_qcd((4, 0), nf))
                    # N3LO
                    if self.order[0] >= 4:
                        b_qcd_vec.append(b_qcd((5, 0), nf))
            a_s = unidimensional_exact(beta0_qcd, b_qcd_vec, u, a_ref[0], "Radau", 1e-6)
            return np.array([a_s, a_em])
        # otherwise rescale the RGE to run in terms of
        # u = ln(scale_to/scale_from)
        beta_qcd_vec = [beta_qcd((2, 0), nf)]
        beta_qcd_mix = 0
        # NLO
        if self.order[0] >= 2:
            beta_qcd_vec.append(beta_qcd((3, 0), nf))
            beta_qed_mix = beta_qed((1, 2), nf)
            # NNLO
            if self.order[0] >= 3:
                beta_qcd_vec.append(beta_qcd((4, 0), nf))
                # N3LO
                if self.order[0] >= 4:
                    beta_qcd_vec.append(beta_qcd((5, 0), nf))
        beta_qed_vec = [beta_qed((0, 2), nf)]
        beta_qed_mix = 0
        if self.order[1] >= 2:
            beta_qcd_mix = beta_qcd((2, 1), nf)
            beta_qed_vec.append(beta_qed((0, 3), nf))
        # integration kernel
        def rge(_t, a, beta_qcd_vec, beta_qed_vec):
            rge_qcd = -(a[0] ** 2) * (
                np.sum([a[0] ** k * b for k, b in enumerate(beta_qcd_vec)])
                + a[1] * beta_qcd_mix
            )
            rge_qed = -(a[1] ** 2) * (
                np.sum([a[1] ** k * b for k, b in enumerate(beta_qed_vec)])
                + a[0] * beta_qed_mix
            )
            res = np.array([rge_qcd, rge_qed])
            return res

        # let scipy solve
        res = scipy.integrate.solve_ivp(
            rge,
            (0, u),
            a_ref,
            args=[beta_qcd_vec, beta_qed_vec],
            method="Radau",
            rtol=1e-6,
        )
        return np.array([res.y[0][-1], res.y[1][-1]])

    def compute(self, a_ref, nf, scale_from, scale_to):
        """Compute actual couplings.

        This is a wrapper in order to pass the computation to the corresponding method
        (depending on the calculation method).

        Parameters
        ----------
        a_ref : numpy.ndarray
            reference a
        nf : int
            value of nf for computing alpha
        scale_from : float
            reference scale
        scale_to : float
            target scale

        Returns
        -------
        numpy.ndarray
            couplings at target scale :math:`a(Q^2)`
        """
        key = (float(a_ref[0]), float(a_ref[1]), nf, scale_from, float(scale_to))
        try:
            return self.cache[key].copy()
        except KeyError:
            # at the moment everything is expanded - and type has been checked in the constructor
            if self.method == "exact":
                a_new = self.compute_exact(
                    a_ref.astype(float), nf, scale_from, scale_to
                )
            else:
                a_new = couplings_expanded(
                    self.order, a_ref.astype(float), nf, scale_from, float(scale_to)
                )
            self.cache[key] = a_new.copy()
            return a_new

    def a(
        self,
        scale_to,
        fact_scale=None,
        nf_to=None,
    ):
        r"""Compute couplings :math:`a_i(\mu_R^2) = \frac{\alpha_i(\mu_R^2)}{4\pi}`.

        Parameters
        ----------
        scale_to : float
            final scale to evolve to :math:`\mu_R^2`
        fact_scale : float
            factorization scale (if different from final scale)

        Returns
        -------
        numpy.ndarray
            couplings :math:`a_i(\mu_R^2) = \frac{\alpha_i(\mu_R^2)}{4\pi}`
        """
        # Set up the path to follow in order to go from q2_0 to q2_ref
        final_a = self.a_ref.copy()
        path = self.thresholds.path(scale_to, nf_to)
        is_downward = thresholds.is_downward_path(path)
        shift = thresholds.flavor_shift(is_downward)

        # as a default assume mu_F^2 = mu_R^2
        if fact_scale is None:
            fact_scale = scale_to
        for k, seg in enumerate(path):
            # skip a very short segment, but keep the matching
            if not np.isclose(seg.q2_from, seg.q2_to):
                new_a = self.compute(final_a, seg.nf, seg.q2_from, seg.q2_to)
            else:
                new_a = final_a
            # apply matching conditions: see hep-ph/9706430
            # - if there is yet a step to go
            if k < len(path) - 1:
                # q2_to is the threshold value
                L = np.log(scale_to / fact_scale) + np.log(
                    self.thresholds.thresholds_ratios[seg.nf - shift]
                )
                m_coeffs = (
                    compute_matching_coeffs_down(self.hqm_scheme, seg.nf - 1)
                    if is_downward
                    else compute_matching_coeffs_up(self.hqm_scheme, seg.nf)
                )
                fact = 1.0
                # shift
                for n in range(1, self.order[0]):
                    for l_pow in range(n + 1):
                        fact += new_a[0] ** n * L**l_pow * m_coeffs[n, l_pow]
                new_a[0] *= fact
            final_a = new_a
        return final_a

    def a_s(self, scale_to, fact_scale=None, nf_to=None):
        r"""Compute coupling :math:`a_s(\mu_R^2) = \frac{\alpha_s(\mu_R^2)}{4\pi}`.

        Parameters
        ----------
        scale_to : float
            final scale to evolve to :math:`\mu_R^2`
        fact_scale : float
            factorization scale (if different from final scale)

        Returns
        -------
        a_s : float
            couplings :math:`a_s(\mu_R^2) = \frac{\alpha_s(\mu_R^2)}{4\pi}`
        """
        return self.a(scale_to, fact_scale, nf_to)[0]


def compute_matching_coeffs_up(mass_scheme, nf):
    r"""Compute the upward matching coefficients.

    The matching coefficients :cite:`Schroder:2005hy,Chetyrkin:2005ia,Vogt:2004ns`
    are normalized at threshold when moving to a regime with *more* flavors.

    We follow notation of :cite:`Vogt:2004ns` (eq 2.43) for POLE masses.

    The *inverse* |MSbar| values instead are given in :cite:`Schroder:2005hy` (eq 3.1)
    multiplied by a factor of 4 (and 4^2 ...).

    .. math::
        a_s^{(n_l+1)} = a_s^{(n_l)} + \sum\limits_{n=1} (a_s^{(n_l)})^n
                                \sum\limits_{k=0}^n c_{nl} \log(\mu_R^2/\mu_F^2)

    Parameters
    ----------
    mass_scheme : str
        Heavy quark mass scheme: "POLE" or "MSBAR"
    nf : int
        number of active flavors in the lower patch

    Returns
    -------
    numpy.ndarray
        forward matching coefficient matrix
    """
    matching_coeffs_up = np.zeros((4, 4))
    if mass_scheme == "MSBAR":
        matching_coeffs_up[2, 0] = -22.0 / 9.0
        matching_coeffs_up[2, 1] = 22.0 / 3.0

        # c30 = -d30
        matching_coeffs_up[3, 0] = -62.2116 + 5.4177 * nf
        # c31 = -d31 + 5 c11 * c20
        matching_coeffs_up[3, 1] = 365.0 / 3.0 - 67.0 / 9.0 * nf
        matching_coeffs_up[3, 2] = 109.0 / 3.0 + 16.0 / 9.0 * nf

    elif mass_scheme == "POLE":
        matching_coeffs_up[2, 0] = 14.0 / 3.0
        matching_coeffs_up[2, 1] = 38.0 / 3.0
        matching_coeffs_up[3, 0] = 340.729 - 16.7981 * nf
        matching_coeffs_up[3, 1] = 8941.0 / 27.0 - 409.0 / 27.0 * nf
        matching_coeffs_up[3, 2] = 511.0 / 9.0

    matching_coeffs_up[1, 1] = 4.0 / 3.0 * constants.TR
    matching_coeffs_up[2, 2] = 4.0 / 9.0
    matching_coeffs_up[3, 3] = 8.0 / 27.0

    return matching_coeffs_up


# inversion of the matching coefficients
def compute_matching_coeffs_down(mass_scheme, nf):
    """Compute the downward matching coefficients.

    This is the inverse function to :meth:`compute_matching_coeffs_up`.

    Parameters
    ----------
    mass_scheme : str
        Heavy quark mass scheme: "POLE" or "MSBAR"
    nf : int
        number of active flavors in the lower patch

    Returns
    -------
    numpy.ndarray
        downward matching coefficient matrix
    """
    c_up = compute_matching_coeffs_up(mass_scheme, nf)
    return invert_matching_coeffs(c_up)


def invert_matching_coeffs(c_up):
    """Compute the perturbative inverse of the matching conditions.

    They can be obtained e.g. via Mathematica by:

    .. code-block:: Mathematica

        Module[{f, g, l, sol},
            f[a_] := a + Sum[d[n, k]*L^k*a^(1 + n), {n, 3}, {k, 0, n}];
            g[a_] := a + Sum[c[n, k]*L^k*a^(1 + n), {n, 3}, {k, 0, n}] /. {c[1, 0] -> 0};
            l = CoefficientList[Normal@Series[f[g[a]], {a, 0, 5}], {a, L}];
            sol = First@
                Solve[{l[[3]] == 0, l[[4]] == 0, l[[5]] == 0},
                Flatten@Table[d[n, k], {n, 3}, {k, 0, n}]];
            Do[Print@r, {r, sol}];
            Print@Series[f[g[a]] /. sol, {a, 0, 5}];
            Print@Series[g[f[a]] /. sol, {a, 0, 5}];
        ]

    Parameters
    ----------
    c_up : numpy.ndarray
        forward matching coefficient matrix

    Returns
    -------
    numpy.ndarray
        downward matching coefficient matrix
    """
    matching_coeffs_down = np.zeros_like(c_up)
    matching_coeffs_down[1, 1] = -c_up[1, 1]
    matching_coeffs_down[2, 0] = -c_up[2, 0]
    matching_coeffs_down[2, 1] = -c_up[2, 1]
    matching_coeffs_down[2, 2] = 2.0 * c_up[1, 1] ** 2 - c_up[2, 2]
    matching_coeffs_down[3, 0] = -c_up[3, 0]
    matching_coeffs_down[3, 1] = 5 * c_up[1, 1] * c_up[2, 0] - c_up[3, 1]
    matching_coeffs_down[3, 2] = 5 * c_up[1, 1] * c_up[2, 1] - c_up[3, 2]
    matching_coeffs_down[3, 3] = (
        -5 * c_up[1, 1] ** 3 + 5 * c_up[1, 1] * c_up[2, 2] - c_up[3, 3]
    )
    return matching_coeffs_down



# -*- coding: utf-8 -*-
"""This module contains the :class:`OperatorGrid` class.

The first is the driver class of eko as it is the one that collects all the
previously instantiated information and does the actual computation of the Q2s.
"""

import logging
import numbers

import numpy as np

from .. import basis_rotation as br
from .. import matching_conditions, member
from .. import scale_variations as sv
from ..matching_conditions.operator_matrix_element import OperatorMatrixElement
from ..thresholds import flavor_shift, is_downward_path
from . import Operator, flavors, physical

logger = logging.getLogger(__name__)


class OperatorGrid(sv.ModeMixin):
    """
    The operator grid is the driver class of the evolution.

    It receives as input a threshold holder and a generator of a_s.
    From that point onwards it can compute any operator at any q2.

    Parameters
    ----------
        config: dict
            configuration dictionary
        q2_grid: array
            Grid in Q2 on where to to compute the operators
        order: tuple(int,int)
            orders in perturbation theory
        thresholds_config: eko.thresholds.ThresholdsAtlas
            Instance of :class:`~eko.thresholds.Threshold` containing information about the
            thresholds
        strong_coupling: eko.strong_coupling.StrongCoupling
            Instance of :class:`~eko.strong_coupling.StrongCoupling` able to generate a_s for
            any q
        kernel_dispatcher: eko.kernel_generation.KernelDispatcher
            Instance of the :class:`~eko.kernel_generation.KernelDispatcher` with the
            information about the kernels
    """

    def __init__(
        self,
        config,
        q2_grid,
        thresholds_config,
        strong_coupling,
        interpol_dispatcher,
    ):
        # check
        order = config["order"]
        method = config["method"]
        if not method in [
            "iterate-exact",
            "iterate-expanded",
            "truncated",
            "ordered-truncated",
            "decompose-exact",
            "decompose-expanded",
            "perturbative-exact",
            "perturbative-expanded",
        ]:
            raise ValueError(f"Unknown evolution mode {method}")
        if order == (1, 0) and method != "iterate-exact":
            logger.warning("Evolution: In LO we use the exact solution always!")

        self.config = config
        self.q2_grid = q2_grid
        self.managers = dict(
            thresholds_config=thresholds_config,
            strong_coupling=strong_coupling,
            interpol_dispatcher=interpol_dispatcher,
        )
        self._threshold_operators = {}
        self._matching_operators = {}

    @classmethod
    def from_dict(
        cls,
        theory_card,
        operators_card,
        thresholds_config,
        strong_coupling,
        interpol_dispatcher,
    ):
        """
        Create the object from the theory dictionary.

        Parameters
        ----------
            theory_card : dict
                theory dictionary
            thresholds_config : eko.thresholds.ThresholdsAtlas
                An instance of the ThresholdsAtlas class
            strong_coupling : eko.strong_coupling.StrongCoupling
                An instance of the StrongCoupling class
            interpol_dispatcher : eko.interpolation.InterpolatorDispatcher
                An instance of the InterpolatorDispatcher class

        Returns
        -------
            obj : cls
                created object
        """
        config = {}
        config["order"] = tuple(int(o) for o in theory_card["order"])
        method = theory_card["ModEv"]
        mod_ev2method = {
            "EXA": "iterate-exact",
            "EXP": "iterate-expanded",
            "TRN": "truncated",
        }
        method = mod_ev2method.get(method, method)
        config["method"] = method
        config["backward_inversion"] = operators_card["backward_inversion"]
        config["fact_to_ren"] = (theory_card["fact_to_ren_scale_ratio"]) ** 2
        config["ev_op_max_order"] = operators_card["ev_op_max_order"]
        config["ev_op_iterations"] = operators_card["ev_op_iterations"]
        config["debug_skip_singlet"] = operators_card["debug_skip_singlet"]
        config["debug_skip_non_singlet"] = operators_card["debug_skip_non_singlet"]
        config["n_integration_cores"] = operators_card["n_integration_cores"]
        config["HQ"] = theory_card["HQ"]
        config["ModSV"] = theory_card["ModSV"]
        q2_grid = np.array(operators_card["Q2grid"], np.float_)
        intrinsic_range = []
        if int(theory_card["IC"]) == 1:
            intrinsic_range.append(4)
        if int(theory_card["IB"]) == 1:
            intrinsic_range.append(5)
        config["intrinsic_range"] = intrinsic_range
        for hq in br.quark_names[3:]:
            config[f"m{hq}"] = theory_card[f"m{hq}"]
        return cls(
            config, q2_grid, thresholds_config, strong_coupling, interpol_dispatcher
        )

    def get_threshold_operators(self, path):
        """
        Generate the threshold operators.

        This method is called everytime the OperatorGrid is asked for a grid on Q^2
        with a list of the relevant areas.
        If new threshold operators need to be computed, they will be
        cached in an internal dictionary.

        The internal dictionary is self._threshold_operators and its structure is:
        (q2_from, q2_to) -> eko.operators.Operator

        It computes and stores the necessary macthing operators
        Parameters
        ----------
            path: list(`eko.thresholds.PathSegment`)
                thresholds path

        Returns
        -------
            thr_ops: list(eko.evolution_operator.Operator)
        """
        # The base area is always that of the reference q
        thr_ops = []
        # is_downward point to smaller nf
        is_downward = is_downward_path(path)
        shift = flavor_shift(is_downward)
        for seg in path[:-1]:
            new_op_key = seg.tuple
            thr_config = self.managers["thresholds_config"]
            kthr = thr_config.thresholds_ratios[seg.nf - shift]
            ome = OperatorMatrixElement(
                self.config,
                self.managers,
                seg.nf - shift + 3,
                seg.q2_to,
                is_downward,
                np.log(kthr),
                self.config["HQ"] == "MSBAR",
            )
            if new_op_key not in self._threshold_operators:
                # Compute the operator and store it
                logger.info("Prepare threshold operator")
                op_th = Operator(
                    self.config,
                    self.managers,
                    seg.nf,
                    seg.q2_from,
                    seg.q2_to,
                    is_threshold=True,
                )
                op_th.compute()
                self._threshold_operators[new_op_key] = op_th
            thr_ops.append(self._threshold_operators[new_op_key])

            # Compute the matching conditions and store it
            if seg.q2_to not in self._matching_operators:
                ome.compute()
                self._matching_operators[seg.q2_to] = ome.op_members
        return thr_ops

    def compute(self, q2grid=None):
        """Computes all ekos for the `q2grid`.

        Parameters
        ----------
        q2grid: list(float)
            List of :math:`Q^2`

        Returns
        -------
        list(dict)
            List of ekos for each value of :math:`Q^2`
        """
        # use input?
        if q2grid is None:
            q2grid = self.q2_grid
        # normalize input
        if isinstance(q2grid, numbers.Number):
            q2grid = [q2grid]
        # And now return the grid
        grid_return = {}
        for q2 in q2grid:
            # shift path for expanded scheme
            q2_gen = (
                q2 * self.config["fact_to_ren"]
                if self.sv_mode == sv.Modes.expanded
                else q2
            )
            grid_return[q2] = self.generate(q2_gen)
        return grid_return

    def generate(self, q2):
        r"""Computes a single EKO.

        Parameters
        ----------
        q2: float
            Target value of :math:`Q^2`

        Returns
        -------
        dict
            eko :math:`\mathbf E(Q^2 \leftarrow Q_0^2)` in flavor basis as numpy array
        """
        # The lists of areas as produced by the thresholds
        path = self.managers["thresholds_config"].path(q2)
        # Prepare the path for the composition of the operator
        thr_ops = self.get_threshold_operators(path)
        # we start composing with the highest operator ...
        operator = Operator(
            self.config, self.managers, path[-1].nf, path[-1].q2_from, path[-1].q2_to
        )
        operator.compute()
        intrinsic_range = self.config["intrinsic_range"]
        is_downward = is_downward_path(path)
        if is_downward:
            intrinsic_range = [4, 5, 6]
        final_op = physical.PhysicalOperator.ad_to_evol_map(
            operator.op_members,
            operator.nf,
            operator.q2_to,
            intrinsic_range,
        )
        # and multiply the lower ones from the right
        for op in reversed(list(thr_ops)):
            phys_op = physical.PhysicalOperator.ad_to_evol_map(
                op.op_members, op.nf, op.q2_to, intrinsic_range
            )

            # join with the basis rotation, since matching requires c+ (or likewise)
            if is_downward:
                matching = matching_conditions.MatchingCondition.split_ad_to_evol_map(
                    self._matching_operators[op.q2_to],
                    op.nf - 1,
                    op.q2_to,
                    intrinsic_range=intrinsic_range,
                )
                invrot = member.ScalarOperator.promote_names(
                    flavors.rotate_matching_inverse(op.nf), op.q2_to
                )
                final_op = final_op @ matching @ invrot @ phys_op
            else:
                matching = matching_conditions.MatchingCondition.split_ad_to_evol_map(
                    self._matching_operators[op.q2_to],
                    op.nf,
                    op.q2_to,
                    intrinsic_range=intrinsic_range,
                )
                rot = member.ScalarOperator.promote_names(
                    flavors.rotate_matching(op.nf + 1), op.q2_to
                )
                final_op = final_op @ rot @ matching @ phys_op

        values, errors = final_op.to_flavor_basis_tensor()
        return {
            "operators": values,
            "operator_errors": errors,
        }



# -*- coding: utf-8 -*-
"""This file contains the leading-order Altarelli-Parisi splitting kernels."""

import numba as nb
import numpy as np

from .. import constants


@nb.njit(cache=True)
def gamma_ns(N, s1):
    """
    Computes the leading-order non-singlet anomalous dimension.

    Implements Eq. (3.4) of :cite:`Moch:2004pa`.

    Parameters
    ----------
      N : complex
        Mellin moment
      s1 : complex
        harmonic sum :math:`S_{1}`

    Returns
    -------
      gamma_ns : complex
        Leading-order non-singlet anomalous dimension :math:`\\gamma_{ns}^{(0)}(N)`
    """
    gamma = -(3.0 - 4.0 * s1 + 2.0 / N / (N + 1.0))
    result = constants.CF * gamma
    return result


@nb.njit(cache=True)
def gamma_qg(N, nf):
    """
    Computes the leading-order quark-gluon anomalous dimension

    Implements Eq. (3.5) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
      N : complex
        Mellin moment
      nf : int
        Number of active flavors

    Returns
    -------
      gamma_qg : complex
        Leading-order quark-gluon anomalous dimension :math:`\\gamma_{qg}^{(0)}(N)`
    """
    gamma = -(N**2 + N + 2.0) / (N * (N + 1.0) * (N + 2.0))
    result = 2.0 * constants.TR * 2.0 * nf * gamma
    return result


@nb.njit(cache=True)
def gamma_gq(N):
    """
    Computes the leading-order gluon-quark anomalous dimension

    Implements Eq. (3.5) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
      N : complex
        Mellin moment

    Returns
    -------
      gamma_gq : complex
        Leading-order gluon-quark anomalous dimension :math:`\\gamma_{gq}^{(0)}(N)`
    """
    gamma = -(N**2 + N + 2.0) / (N * (N + 1.0) * (N - 1.0))
    result = 2.0 * constants.CF * gamma
    return result


@nb.njit(cache=True)
def gamma_gg(N, s1, nf):
    """
    Computes the leading-order gluon-gluon anomalous dimension

    Implements Eq. (3.5) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
      N : complex
        Mellin moment
      s1 : complex
        harmonic sum :math:`S_{1}`
      nf : int
        Number of active flavors

    Returns
    -------
      gamma_gg : complex
        Leading-order gluon-gluon anomalous dimension :math:`\\gamma_{gg}^{(0)}(N)`
    """
    gamma = s1 - 1.0 / N / (N - 1.0) - 1.0 / (N + 1.0) / (N + 2.0)
    result = constants.CA * (4.0 * gamma - 11.0 / 3.0) + 4.0 / 3.0 * constants.TR * nf
    return result


@nb.njit(cache=True)
def gamma_singlet(N, s1, nf):
    r"""
      Computes the leading-order singlet anomalous dimension matrix

      .. math::
          \gamma_S^{(0)} = \left(\begin{array}{cc}
            \gamma_{qq}^{(0)} & \gamma_{qg}^{(0)}\\
            \gamma_{gq}^{(0)} & \gamma_{gg}^{(0)}
          \end{array}\right)

      Parameters
      ----------
        N : complex
          Mellin moment
        s1 : complex
          harmonic sum :math:`S_{1}`
        nf : int
          Number of active flavors

      Returns
      -------
        gamma_S_0 : numpy.ndarray
          Leading-order singlet anomalous dimension matrix :math:`\gamma_{S}^{(0)}(N)`

      See Also
      --------
        gamma_ns : :math:`\gamma_{qq}^{(0)}`
        gamma_qg : :math:`\gamma_{qg}^{(0)}`
        gamma_gq : :math:`\gamma_{gq}^{(0)}`
        gamma_gg : :math:`\gamma_{gg}^{(0)}`
    """
    gamma_qq = gamma_ns(N, s1)
    gamma_S_0 = np.array(
        [[gamma_qq, gamma_qg(N, nf)], [gamma_gq(N), gamma_gg(N, s1, nf)]], np.complex_
    )
    return gamma_S_0
